# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from types import SimpleNamespace

import numpy as np

from ._scheme import Scheme


class Tetrahedron(Scheme):
    r"""A quadrature scheme suitable for Tetrahedrons of order 1, 2 or 3 on the interval
    :math:`[0, 1]`.

    Parameters
    ----------
    order : int
        The order of the quadrature scheme.

    Notes
    -----

    The approximation is given by

    ..  math::

        \int_V f(x) dV \approx \sum f(x_q) w_q

    with quadrature points :math:`x_q` and corresponding weights :math:`w_q` [1]_.

    References
    ----------
    .. [1] O. C. Zienkiewicz, R. L. Taylor and J. Z. Zhu, The Finite Element Method: Its
       Basis and Fundamentals, 7th ed., Elsevier, 2013.
    """

    def __init__(self, order: int):
        scheme = SimpleNamespace()

        if order == 1:
            scheme.points = np.ones((1, 3)) / 4
            scheme.weights = np.ones(1) / 6

        elif order == 2:
            a = 0.13819660
            b = 0.58541020
            scheme.points = np.array([[a, a, a], [b, a, a], [a, b, a], [a, a, b]])
            scheme.weights = np.ones(4) / 24

        elif order == 3:
            a = 1 / 6
            b = 1 / 2
            c = 1 / 4
            d = 9 / 120
            scheme.points = np.array(
                [[a, a, a], [b, a, a], [a, b, a], [b, b, a], [c, c, c]]
            )
            scheme.weights = np.array([d, d, d, d, -4 / 30])

        elif order == 5:
            a = 1 / 2
            b = 0.6984197043243866
            c = 0.1005267652252045
            d = 0.0568813795204234
            e = 0.31437287349319221
            f = 0.019047619047619
            g = 0.0885898247429807
            h = 0.1328387466855907
            scheme.points = np.array(
                [
                    [0, a, a, a, 0, 0, b, c, c, c, d, e, e, e],
                    [a, 0, a, 0, a, 0, c, c, c, b, e, e, e, d],
                    [a, a, 0, 0, 0, a, c, c, b, c, e, e, d, e],
                ]
            ).T
            scheme.weights = np.array([f, f, f, f, f, f, g, g, g, g, h, h, h, h]) / 6

        else:
            raise NotImplementedError("order must be either 1, 2, 3 or 5.")

        super().__init__(scheme.points, scheme.weights)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ..tools import force
from ._job import Job


class CharacteristicCurve(Job):
    r"""A job with a list of steps and a method to evaluate them. Force-displacement
    curve data is tracked during evaluation for a given :class:`~felupe.Boundary` by
    a built-in ``callback``.

    Parameters
    ----------
    steps : list of Step
        A list with steps, where each step subsequently depends on the solution of the
        previous step.
    items : list of SolidBody, SolidBodyNearlyIncompressible, SolidBodyPressure, SolidBodyGravity, PointLoad, MultiPointConstraint, MultiPointContact or None, optional
        A list of items with methods for the assembly of sparse vectors/matrices which
        are used to evaluate the sum of reaction forces. If None, the total reaction
        forces from the :class:`~felupe.tools.NewtonResult` of the substep are used.
    callback : callable, optional
        A callable which is called after each completed substep. Function signature must
        be ``lambda stepnumber, substepnumber, substep, **kwargs: None``, where
        ``substep`` is an instance of :class:`~felupe.tools.NewtonResult`. The field
        container of the completed substep is available as ``substep.x``. Default
        is ``callback=lambda stepnumber, substepnumber, substep, **kwargs: None``.
    **kwargs : dict, optional
        Optional keyword-arguments for the ``callback`` function.

    Examples
    --------
    >>> import felupe as fem
    >>>
    >>> mesh = fem.Cube(n=6)
    >>> region = fem.RegionHexahedron(mesh)
    >>> field = fem.FieldContainer([fem.Field(region, dim=3)])
    >>>
    >>> boundaries = dict()
    >>> boundaries["fixed"] = fem.Boundary(field[0], fx=0, skip=(False, False, False))
    >>> boundaries["clamped"] = fem.Boundary(field[0], fx=1, skip=(True, False, False))
    >>> boundaries["move"] = fem.Boundary(field[0], fx=1, skip=(False, True, True))
    >>>
    >>> umat = fem.NeoHooke(mu=1, bulk=2)
    >>> solid = fem.SolidBody(umat, field)
    >>>
    >>> move = fem.math.linsteps([0, 1], num=5)
    >>> step = fem.Step(items=[solid], ramp={boundaries["move"]: move}, boundaries=boundaries)
    >>>
    >>> job = fem.CharacteristicCurve(steps=[step], boundary=boundaries["move"])
    >>> job.evaluate()
    >>> fig, ax = job.plot(
    >>>    xlabel=r"Displacement $u_1$ in mm $\rightarrow$",
    >>>    ylabel=r"Normal Force in $F_1$ in N $\rightarrow$",
    >>>    marker="o",
    >>> )
    >>> ax2 = solid.imshow("Principal Values of Cauchy Stress")

    See Also
    --------
    Step : A Step with multiple substeps, subsequently depending on the solution
        of the previous substep.
    Job : A job with a list of steps and a method to evaluate them.
    tools.NewtonResult : A data class which represents the result found by
        Newton's method.

    """

    def __init__(
        self,
        steps,
        boundary,
        items=None,
        callback=lambda stepnumber, substepnumber, substep, **kwargs: None,
        **kwargs,
    ):
        super().__init__(steps, self._callback, **kwargs)

        self.items = items
        self.boundary = boundary
        self.x = []
        self.y = []
        self.res = None
        self._cb = callback

    def _callback(self, stepnumber, substepnumber, substep, **kwargs):
        if self.items is not None:
            fun = sum([item.results.force for item in self.items])
        else:
            fun = substep.fun

        self.x.append(substep.x[0].values[self.boundary.points[0]])
        self.y.append(force(substep.x, fun, self.boundary))
        self.res = substep

        self._cb(stepnumber, substepnumber, substep, **kwargs)

    def plot(
        self,
        x=None,
        y=None,
        xaxis=0,
        yaxis=0,
        xlabel=None,
        ylabel=None,
        xscale=1.0,
        yscale=1.0,
        gradient=False,
        swapaxes=False,
        ax=None,
        items=None,
        **kwargs,
    ):
        """Plot force-displacement characteristic curves on a pre-evaluated job,
        tracked on a given :class:`~felupe.Boundary`.

        Parameters
        ----------
        x : list of ndarray or None, optional
            A list with arrays of displacement data. If None, the displacement is taken
            from the first field of the field container from each completed substep. The
            displacement data is then taken from the first point of the tracked
            :class:`~felupe.Boundary`. Default is None.
        y : list of ndarray or None, optional
            A list with arrays of reaction force data. If None, the force is taken
            from the :class:`~felupe.tools.NewtonResult` of each completed substep.
            Default is None.
        xaxis : int, optional
            The axis for the displacement data (default is 0).
        yaxis : int, optional
            The axis for the reaction force data (default is 0).
        xlabel : str or None, optional
            The label of the x-axis (default is None).
        ylabel : str or None, optional
            The label of the y-axis (default is None).
        xscale : float, optional
            A scaling factor for the displacement data (default is 1.0).
        yscale : float, optional
            A scaling factor the reaction force data (default is 1.0).
        gradient : bool, optional
            A flag to plot the gradient of the y-data. Uses
            ``numpy.gradient(edge_order=2)``. The gradient data is set to ``np.nan`` for
            absolute values greater than the mean value plus two times the standard
            deviation. Default is False.
        swapaxes : bool, optional
            A flag to flip the plot (x, y) to (y, x). Also changes the labels.
        ax : matplotlib.axes.Axes
            An axes object where the plot is placed in.
        items : slice, ndarray or None
            Indices or a range of data points to plot. If None, all data points are
            plotted (default is None).
        **kwargs : dict
            Additional keyword arguments for plotting in ``ax.plot(**kwags)``.

        Returns
        -------
        fig : matplotlib.figure.Figure
            The figure object where the plot is placed in.
        ax : matplotlib.axes.Axes
            The axes object where the plot is placed in.

        """

        if self.res is None:
            raise ValueError(
                "Results are empty. Run `job.evaluate()` and call `job.plot()` again."
            )

        import matplotlib.pyplot as plt

        if x is None:
            x = self.x
        if y is None:
            y = self.y

        if items is None:
            items = slice(None)

        x = np.array(x)[items]
        y = np.array(y)[items]

        if gradient:
            y = np.gradient(y, x[:, xaxis], edge_order=2, axis=0)
            z = np.gradient(y, x[:, xaxis], edge_order=2, axis=0)
            cuttoff = np.mean(abs(z[:, yaxis])) + 2 * np.std(abs(z[:, yaxis]))
            y[abs(z) > cuttoff] = np.nan

        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.get_figure()

        if swapaxes:
            x, y = y, x
            xlabel, ylabel = ylabel, xlabel
            xaxis, yaxis = yaxis, xaxis
            xscale, yscale = yscale, xscale

        ax.plot(x[:, xaxis] * xscale, y[:, yaxis] * yscale, **kwargs)

        if xlabel is not None:
            ax.set_xlabel(xlabel)

        if ylabel is not None:
            ax.set_ylabel(ylabel)

        return fig, ax



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np

from ..assembly import IntegralForm
from ..constitution import AreaChange
from ..field import FieldAxisymmetric
from ..math import det


class Assemble:
    "A class with assembly methods of a SolidBody."

    def __init__(self, vector, matrix):
        self.vector = vector
        self.matrix = matrix


class Evaluate:
    "A class with evaluate methods of a SolidBody."

    def __init__(self, gradient, hessian, cauchy_stress=None, kirchhoff_stress=None):
        self.gradient = gradient
        self.hessian = hessian

        if cauchy_stress is not None:
            self.cauchy_stress = cauchy_stress
            self.kirchhoff_stress = kirchhoff_stress


class Results:
    "A class with intermediate results of a SolidBody."

    def __init__(self, stress=False, elasticity=False):
        self.force = None
        self._force = None
        self.stiffness = None
        self.kinematics = None
        self.statevars = None
        self._statevars = None

        self.gradient = None
        self.hessian = None

        self.force_values = None
        self.stiffness_values = None

        self._force_values = None
        self._stiffness_values = None

        if stress:
            self.stress = None

        if elasticity:
            self.elasticity = None

    def update_statevars(self):
        if self._statevars is not None:
            self.statevars = self._statevars


class StateNearlyIncompressible:
    r"""A State with internal cell-wise constant dual fields for (nearly) incompressible
    solid bodies.

    Notes
    -----
    The internal fields :math:`p` and :math:`\bar{J}` are treated as state variables,
    directly derived from the displacement field. Hence, these dual fields are not
    exported to the global degrees of freedom.

    Parameters
    ----------
    field : FieldContainer
        A field container with the displacement field.

    See Also
    --------
    felupe.SolidBodyNearlyIncompressible : A (nearly) incompressible solid body with
        methods for the assembly of sparse vectors/matrices.
    """

    def __init__(self, field):
        self.field = field
        self.dJdF = AreaChange().function
        self.v = None

        # initial values (on mesh-points) of the displacement field
        self.u = field[0].values

        # deformation gradient
        self.F = field.extract()
        self.detF = None

        # cell-values of the internal pressure and volume-ratio fields
        self.p = np.zeros(field.region.mesh.ncells)
        self.J = np.ones(field.region.mesh.ncells)

    def integrate_shape_function_gradient(self, parallel=False, out=None):
        r"""Integrated sub-block matrix containing the shape-functions gradient w.r.t.
        the deformed coordinates :math:`\boldsymbol{K}_{\boldsymbol{u}p}`.

        ..  math::

            \int_V \delta \boldsymbol{F} : \frac{\partial J}{\partial \boldsymbol{F}}
                ~ dV\ \Delta p
            \longrightarrow \boldsymbol{K}_{\boldsymbol{u}p}

        """

        return IntegralForm(
            fun=self.dJdF(self.F), v=self.field, dV=self.field.region.dV
        ).integrate(parallel=parallel, out=out)[0]

    def volume(self):
        r"""Return the cell volumes of the deformed configuration.

        ..  math::

            v = \int_V J ~ dV

        """
        dV = self.field.region.dV
        if isinstance(self.field[0], FieldAxisymmetric):
            R = self.field[0].radius
            dA = self.field.region.dV
            dV = 2 * np.pi * R * dA
        dv = det(self.F[0]) * dV
        return dv.sum(0, out=self.v)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import inspect

import numpy as np

from ..math import cdya_ik, cdya_il, ddot, det, dya, identity, inv, transpose
from ._base import ConstitutiveMaterial


class NearlyIncompressible(ConstitutiveMaterial):
    r"""A nearly-incompressible material formulation to augment the distortional part of
    the strain energy function by a volumetric part and a constraint equation.

    Notes
    -----
    The total potential energy of internal forces is given in Eq.
    :eq:`nearlyincompressible`.

    ..  math::
        :label: nearlyincompressible

        \Pi_{int}(\boldsymbol{F}, p, \bar{J}) =
            \int_V \hat{\psi}(\boldsymbol{F})\ dV +
            \int_V U(\bar{J})\ dV +
            \int_V p (J - \bar{J})\ dV

    The volumetric part of the strain energy density function is denoted in Eq.
    :eq:`nearlyincompressible-volumetric` along with its first and second derivatives.

    ..  math::
        :label: nearlyincompressible-volumetric

        \bar{U} &= \frac{K}{2} \left( \bar{J} - 1 \right)^2

        \bar{U}' &= K \left( \bar{J} - 1 \right)

        \bar{U}'' &= K

    Parameters
    ----------
    material : ConstitutiveMaterial
        A hyperelastic material definition for the distortional part of the strain
        energy density function :math:`\hat{\psi}(\boldsymbol{F})` with methods for the
        ``gradient`` :math:`\partial_\boldsymbol{F}(\hat{\psi})` and the ``hessian``
        :math:`\partial_\boldsymbol{F}[\partial_\boldsymbol{F}(\hat{\psi})]` w.r.t the
        deformation gradient tensor :math:`\boldsymbol{F}`.
    bulk : float
        The bulk modulus :math:`K` for the volumetric part of the strain energy
        function.
    parallel : bool, optional
        A flag to invoke parallel (threaded) math operations (default is False).
    dUdJ : callable, optional
        A function which evaluates the derivative of the volumetric part of the strain
        energy function :math:`\bar{U}'` w.r.t. the volume ratio :math:`\bar{J}`.
        Function signature must be ``lambda J, bulk: dUdJ``. Default is
        :math:`\bar{U}' = K (\bar{J} - 1)` or ``lambda J, bulk: bulk * (J - 1)``.
    d2UdJdJ : callable, optional
        A function which evaluates the second derivative of the volumetric part of the
        strain energy function :math:`\bar{U}''` w.r.t. the volume ratio
        :math:`\bar{J}`. Function signature must be ``lambda J, bulk: d2UdJdJ``.
        Default is :math:`\bar{U}'' = K` or ``lambda J, bulk: bulk``.

    Examples
    --------
    >>> import felupe as fem
    >>>
    >>> field = fem.FieldsMixed(fem.RegionHexahedron(fem.Cube(n=6)), n=3)
    >>> boundaries, loadcase = fem.dof.uniaxial(field, clamped=True)
    >>> umat = fem.NearlyIncompressible(fem.NeoHooke(mu=1), bulk=5000)
    >>> solid = fem.SolidBody(umat, field)
    >>> job = fem.Job(steps=[fem.Step(items=[solid], boundaries=boundaries)]).evaluate()

    See Also
    --------
    ThreeFieldVariation : Hu-Washizu hydrostatic-volumetric selective three-field
        variation for nearly-incompressible material formulations.

    """

    def __init__(
        self,
        material,
        bulk,
        parallel=False,
        dUdJ=lambda J, bulk: bulk * (J - 1),
        d2UdJdJ=lambda J, bulk: bulk,
    ):
        self.material = self.fun = material
        self.parallel = parallel
        self.bulk = bulk
        self.dUdJ = dUdJ
        self.d2UdJdJ = d2UdJdJ
        self.x = [material.x[0], np.ones(1), np.ones(1), material.x[-1]]

    def gradient(self, x, out=None):
        r"""Return a list with the gradient of the strain energy density function
        w.r.t. the fields displacements, pressure and volume ratio.

        Parameters
        ----------
        x : list of ndarray
            List of extracted field values with the deformation gradient tensor
            :math:`\boldsymbol{F}` as first, the pressure :math:`p` as
            second and the volume ratio :math:`\bar{J}` as third list item. Initial
            state variables are stored in the last (fourth) list item.
        out : ndarray or None, optional
            A location into which the result is stored (default is None).

        Returns
        -------
        list of ndarrays
            List of gradients w.r.t. the input variables :math:`\boldsymbol{F}`,
            :math:`p` and :math:`\bar{J}`. The last item of the list contains the
            updated state variables.

        Notes
        -----
        ..  math::

            \delta_\boldsymbol{u}(\Pi_{int}) &=
                \int_V \left( \frac{\partial \hat{\psi}}{\partial \boldsymbol{F}} +
                p\ J \boldsymbol{F}^{-T} \right) : \delta\boldsymbol{F}\ dV

            \delta_p(\Pi_{int}) &=
                \int_V \left( J - \bar{J} \right)\ \delta p\ dV

            \delta_\bar{J}(\Pi_{int}) &=
                \int_V \left( \bar{U}' - p \right)\ \delta \bar{J}\ dV

        """
        kwargs = {}
        if "out" in inspect.signature(self.material.gradient).parameters:
            kwargs["out"] = out

        [F, p, J], statevars = x[:3], x[-1]
        dWdF, statevars_new = self.material.gradient([F, statevars], **kwargs)
        dWdF += p * transpose(inv(F, determinant=1.0))
        dWdp = det(F) - J
        dWdJ = self.dUdJ(J, self.bulk) - p
        return [dWdF, dWdp, dWdJ, statevars_new]

    def hessian(self, x, out=None):
        r"""Return a list with the hessian of the strain energy density function
        w.r.t. the fields displacements, pressure and volume ratio.

        Parameters
        ----------
        x : list of ndarray
            List of extracted field values with the deformation gradient tensor
            :math:`\boldsymbol{F}` as first, the pressure :math:`p` as
            second and the volume ratio :math:`\bar{J}` as third list item. Initial
            state variables are stored in the last (fourth) list item.
        out : ndarray or None, optional
            A location into which the result is stored (default is None).

        Returns
        -------
        list of ndarrays
            List of the hessian w.r.t. the input variables :math:`\boldsymbol{F}`,
            :math:`p` and :math:`\bar{J}`. The upper-triangle items of the hessian are
            returned as the items of the list.

        Notes
        -----
        ..  math::

            \Delta_\boldsymbol{u}\delta_\boldsymbol{u}(\Pi_{int}) &= \int_V
                \delta\boldsymbol{F} : \left[
                \frac{\partial^2 \hat{\psi}}
                {\partial\boldsymbol{F}\ \partial\boldsymbol{F}} +
                p\ J \left( \boldsymbol{F}^{-T} \otimes \boldsymbol{F}^{-T} -
                \boldsymbol{F}^{-T} \overset{il}{\odot} \boldsymbol{F}^{-T} \right)
                \right] : \Delta\boldsymbol{F}\ dV

            \Delta_p\delta_\boldsymbol{u}(\Pi_{int}) &= \int_V
                \delta\boldsymbol{F} : J \boldsymbol{F}^{-T}\ \Delta p\ dV

            \Delta_\bar{J}\delta_\boldsymbol{u}(\Pi_{int}) &= \int_V
                \delta\boldsymbol{F} : \boldsymbol{0}\ \Delta \bar{J}\ dV

            \Delta_p\delta_p(\Pi_{int}) &= \int_V \delta p\ (0)\ \Delta p\ dV

            \Delta_p\delta_\bar{J}(\Pi_{int}) &= \int_V
                \delta \bar{J}\ (-1)\ \Delta p\ dV

            \Delta_\bar{J}\delta_\bar{J}(\Pi_{int}) &= \int_V
                \delta \bar{J}\ \bar{U}''\ \Delta \bar{J}\ dV

        """
        kwargs = {}
        if "out" in inspect.signature(self.material.hessian).parameters:
            kwargs["out"] = out

        [F, p, J], statevars = x[:3], x[-1]
        detF = det(F)
        iFT = transpose(inv(F, determinant=detF))
        d2WdFdF = self.material.hessian([F, statevars], **kwargs)[0]
        d2WdFdF += p * detF * (dya(iFT, iFT) - cdya_il(iFT, iFT))
        d2WdFdp = detF * iFT
        d2WdFdJ = np.zeros_like(F)
        d2Wdpdp = np.zeros_like(p)
        d2WdpdJ = -np.ones_like(J)
        d2WdJdJ = self.d2UdJdJ(J, self.bulk) * np.ones_like(J)
        return [d2WdFdF, d2WdFdp, d2WdFdJ, d2Wdpdp, d2WdpdJ, d2WdJdJ]


class ThreeFieldVariation(ConstitutiveMaterial):
    r"""Hu-Washizu hydrostatic-volumetric selective
    :math:`(\boldsymbol{u},p,J)` - three-field variation for nearly-
    incompressible material formulations. The total potential energy
    for nearly-incompressible hyperelasticity is formulated with a
    determinant-modified deformation gradient. Pressure and volume ratio fields
    should be kept one order lower than the interpolation order of the
    displacement field, e.g. linear displacement fields should be paired with
    element-constant (mean) values of pressure and volume ratio.

    The total potential energy of internal forces is defined with a strain
    energy density function in terms of a determinant-modified deformation
    gradient and an additional control equation.

    ..  math::

        \Pi &= \Pi_{int} + \Pi_{ext}

        \Pi_{int} &= \int_V \psi(\boldsymbol{F}) \ dV \qquad \rightarrow
        \qquad \Pi_{int}(\boldsymbol{u},p,J) = \int_V \psi(\overline{\boldsymbol{F}})
        \ dV + \int_V p (J-\overline{J}) \ dV

        \overline{\boldsymbol{F}} &=
        \left(\frac{\overline{J}}{J}\right)^{1/3} \boldsymbol{F}

    The variations of the total potential energy w.r.t.
    :math:`(\boldsymbol{u},p,J)` lead to the following expressions. We denote
    first partial derivatives as :math:`\boldsymbol{f}_{(\bullet)}` and second
    partial derivatives as :math:`\boldsymbol{A}_{(\bullet,\bullet)}`.

    ..  math::

        \delta_{\boldsymbol{u}} \Pi_{int} &= \int_V \boldsymbol{P} :
        \delta \boldsymbol{F} \ dV = \int_V \left( \frac{\partial \psi}
        {\partial \overline{\boldsymbol{F}}} : \frac{\partial \overline{\boldsymbol{F}}}
        {\partial \boldsymbol{F}} + p J \boldsymbol{F}^{-T} \right) :
        \delta \boldsymbol{F} \ dV

        \delta_{p} \Pi_{int} &= \int_V f_{p} \ \delta p \ dV =
        \int_V (J - \overline{J}) \ \delta p \ dV

        \delta_{\overline{J}} \Pi_{int} &= \int_V f_{\overline{J}} \ \delta \overline{J}
        \ dV = \int_V \left( \frac{\partial \psi}{\partial \overline{\boldsymbol{F}}} :
        \frac{\partial \overline{\boldsymbol{F}}}{\partial \overline{J}} - p \right) :
        \delta \overline{J} \ dV

    The projection tensors from the variations lead the following results.

    ..  math::

        \frac{\partial \overline{\boldsymbol{F}}}{\partial \boldsymbol{F}} &=
        \left(\frac{\overline{J}}{J}\right)^{1/3} \left( \boldsymbol{I}
        \overset{ik}{\odot} \boldsymbol{I} - \frac{1}{3} \boldsymbol{F} \otimes
        \boldsymbol{F}^{-T} \right)

        \frac{\partial \overline{\boldsymbol{F}}}{\partial \overline{J}} &=
        \frac{1}{3 \overline{J}} \overline{\boldsymbol{F}}

    The double-dot products from the variations are now evaluated.

    ..  math::

        \overline{\boldsymbol{P}} &= \frac{\partial \psi}{\partial
        \overline{\boldsymbol{F}}} = \overline{\overline{\boldsymbol{P}}} - \frac{1}{3}
        \left(  \overline{\overline{\boldsymbol{P}}} : \boldsymbol{F} \right)
        \boldsymbol{F}^{-T} \qquad \text{with} \qquad
        \overline{\overline{\boldsymbol{P}}} = \left(\frac{\overline{J}}{J}\right)^{1/3}
        \frac{\partial \psi}{\partial \overline{\boldsymbol{F}}}

        \frac{\partial \psi}{\partial \overline{\boldsymbol{F}}} :
        \frac{1}{3 \overline{J}} \overline{\boldsymbol{F}} &= \frac{1}{3 \overline{J}}
        \overline{\overline{\boldsymbol{P}}} : \boldsymbol{F}

    We now have three formulas; one for the first Piola Kirchhoff stress and
    two additional control equations.

    ..  math::

        \boldsymbol{P} &=
        \overline{\overline{\boldsymbol{P}}} - \frac{1}{3} \left(
        \overline{\overline{\boldsymbol{P}}} : \boldsymbol{F} \right)
        \boldsymbol{F}^{-T}

        f_p &= J - \overline{J}

        f_{\overline{J}} &=  \frac{1}{3 \overline{J}} \left(
        \overline{\overline{\boldsymbol{P}}} : \boldsymbol{F} \right) - p

    A linearization of the above formulas gives six equations (only results are
    given here).

    ..  math::

        \mathbb{A}_{\boldsymbol{u},\boldsymbol{u}} &=  \overline{\overline{\mathbb{A}}}
        + \frac{1}{9} \left(  \boldsymbol{F} : \overline{\overline{\mathbb{A}}} :
        \boldsymbol{F} \right) \boldsymbol{F}^{-T} \otimes \boldsymbol{F}^{-T} -
        \frac{1}{3} \left( \boldsymbol{F}^{-T} \otimes \left(
        \overline{\overline{\boldsymbol{P}}} + \boldsymbol{F} :
        \overline{\overline{\mathbb{A}}} \right) + \left(
        \overline{\overline{\boldsymbol{P}}} + \overline{\overline{\mathbb{A}}} :
        \boldsymbol{F} \right) \otimes \boldsymbol{F}^{-T} \right)

        &+\left( p J + \frac{1}{9} \overline{\overline{\boldsymbol{P}}} : \boldsymbol{F}
        \right) \boldsymbol{F}^{-T} \otimes \boldsymbol{F}^{-T} - \left( p J -
        \frac{1}{3} \overline{\overline{\boldsymbol{P}}} : \boldsymbol{F} \right)
        \boldsymbol{F}^{-T} \overset{il}{\odot} \boldsymbol{F}^{-T}

        A_{p,p} &= 0

        A_{\overline{J},\overline{J}} &= \frac{1}{9 \overline{J}^2} \left(
        \boldsymbol{F} : \overline{\overline{\mathbb{A}}} : \boldsymbol{F} \right) -
        2 \left( \overline{\overline{\boldsymbol{P}}} : \boldsymbol{F} \right)

        \boldsymbol{A}_{\boldsymbol{u},p} &= \boldsymbol{A}_{p, \boldsymbol{u}} =
        J \boldsymbol{F}^{-T}

        \boldsymbol{A}_{\boldsymbol{u},\overline{J}} &= \boldsymbol{A}_{\overline{J},
        \boldsymbol{u}} = \frac{1}{3 \overline{J}} \left( \boldsymbol{P}' +
        \boldsymbol{F} : \overline{\overline{\mathbb{A}}} - \frac{1}{3} \left(
        \boldsymbol{F} : \overline{\overline{\mathbb{A}}} : \boldsymbol{F} \right)
        \boldsymbol{F}^{-T} \right)

        A_{p,\overline{J}} &= A_{\overline{J}, p} = -1


    with

    ..  math::

        \overline{\overline{\mathbb{A}}} = \left(\frac{\overline{J}}{J}\right)^{1/3}
        \frac{\partial^2 \psi}{\partial \overline{\boldsymbol{F}} \partial
        \overline{\boldsymbol{F}}} \left(\frac{\overline{J}}{J}\right)^{1/3}

    as well as

    ..  math::

        \boldsymbol{P}' = \boldsymbol{P} - p J \boldsymbol{F}^{-T}

    Parameters
    ----------
    material : ConstitutiveMaterial
        A hyperelastic material definition for the strain energy density function with
        methods for the ``gradient`` and the ``hessian`` w.r.t the deformation gradient
        tensor.
    parallel : bool, optional
        A flag to invoke parallel (threaded) math operations (default is False).

    """

    def __init__(self, material, parallel=False):
        self.material = self.fun = material

        self._fun_P = self.material.gradient
        self._fun_A = self.material.hessian

        # initial variables for calling
        # ``self.gradient(self.x)`` and ``self.hessian(self.x)``
        self.x = [material.x[0], np.ones(1), np.ones(1), material.x[-1]]

        self.parallel = parallel

    def _gradient_u(self, F, p, J):
        """Variation of total potential w.r.t displacements
        (1st Piola Kirchhoff stress).

        ..  code-block::

            δ_u(Π_int) = ∫_V (∂ψ/∂F + p cof(F)) : δF dV

        Arguments
        ---------
        F : ndarray
            Deformation gradient
        p : ndarray
            Hydrostatic pressure
        J : ndarray
            Volume ratio

        Returns
        -------
        ndarray
            Gradient w.r.t. the deformation gradient

        """

        return self._Pbb - self._PbbF / 3 * self._iFT + p * self._detF * self._iFT

    def _gradient_p(self, F, p, J):
        """Variation of total potential energy w.r.t pressure.

        ..  code-block::

            δ_p(Π_int) = ∫_V (det(F) - J) δp dV

        Arguments
        ---------
        F : ndarray
            Deformation gradient
        p : ndarray
            Hydrostatic pressure
        J : ndarray
            Volume ratio

        Returns
        -------
        ndarray
            Gradient w.r.t. the pressure

        """

        return self._detF - J

    def _gradient_J(self, F, p, J):
        """Variation of total potential energy w.r.t volume ratio.

        ..  code-block::

            δ_J(Π_int) = ∫_V (∂U/∂J - p) δJ dV

        Arguments
        ---------
        F : ndarray
            Deformation gradient
        p : ndarray
            Hydrostatic pressure
        J : ndarray
            Volume ratio

        Returns
        -------
        ndarray
            Gradient w.r.t. the volume ratio

        """

        return self._PbbF / (3 * J) - p

    def gradient(self, x):
        r"""Return a list of variations of the total potential energy w.r.t. the fields
        displacements, pressure and volume ratio.

        Parameters
        ----------
        x : list of ndarray
            List of extracted field values with the Deformation gradient tensor
            :math:`\boldsymbol{F}` as first, the hydrostatic pressure :math:`p` as
            second and the volume ratio :math:`\bar{J}` as third list item.

        Returns
        -------
        list of ndarrays
            List of gradients w.r.t. the input variables :math:`\boldsymbol{F}`,
            :math:`p` and :math:`\bar{J}`.

        """
        kwargs = {}

        [F, p, J], statevars = x[:3], x[-1]

        self._detF = det(F)
        self._iFT = transpose(inv(F))
        self._Fb = (J / self._detF) ** (1 / 3) * F
        self._Pb, statevars_new = self._fun_P([self._Fb, statevars])
        self._Pbb = (J / self._detF) ** (1 / 3) * self._Pb
        self._PbbF = ddot(self._Pbb, F, mode=(2, 2), parallel=self.parallel)

        return [
            self._gradient_u(F, p, J),
            self._gradient_p(F, p, J),
            self._gradient_J(F, p, J),
            statevars_new,
        ]

    def hessian(self, x):
        r"""List of linearized variations of total potential energy w.r.t
        displacements, pressure and volume ratio (these expressions are
        symmetric; ``A_up = A_pu`` if derived from a total potential energy
        formulation). List entries have to be arranged as a flattened list
        from the upper triangle blocks:

        ..  code-block::

            Δ_u(δ_u(Π_int)) = ∫_V δF : (∂²ψ/(∂F∂F) + p ∂cof(F)/∂F) : ΔF dV
            Δ_p(δ_u(Π_int)) = ∫_V δF : J cof(F) Δp dV
            Δ_J(δ_u(Π_int)) = ∫_V δF :  ∂²ψ/(∂F∂J) ΔJ dV
            Δ_p(δ_p(Π_int)) = ∫_V δp 0 Δp dV
            Δ_J(δ_p(Π_int)) = ∫_V δp (-1) ΔJ dV
            Δ_J(δ_J(Π_int)) = ∫_V δJ ∂²ψ/(∂J∂J) ΔJ dV

            [[0 1 2],
             [  3 4],
             [    5]] --> [0 1 2 3 4 5]


        Arguments
        ---------
        extract : list of ndarray
            List of extracted field values with Deformation gradient ``F``
            as first, the hydrostatic pressure ``p`` as second and the
            volume ratio ``J`` as third item.

        Returns
        -------
        list of ndarrays
            List of hessians in upper triangle order

        """

        [F, p, J], statevars = x[:3], x[-1]

        self._detF = det(F)
        self._iFT = transpose(inv(F))
        self._Fb = (J / self._detF) ** (1 / 3) * F
        self._Pbb = (J / self._detF) ** (1 / 3) * self._fun_P([self._Fb, statevars])[0]

        self._eye = identity(F)
        self._P4 = cdya_ik(self._eye, self._eye, parallel=self.parallel) - 1 / 3 * dya(
            F, self._iFT, parallel=self.parallel
        )
        self._A4b = self._fun_A([self._Fb, statevars])[0]
        self._A4bb = (J / self._detF) ** (2 / 3) * self._A4b

        self._PbbF = ddot(self._Pbb, F, mode=(2, 2), parallel=self.parallel)
        self._FA4bb = ddot(F, self._A4bb, mode=(2, 4), parallel=self.parallel)
        self._A4bbF = ddot(self._A4bb, F, mode=(4, 2), parallel=self.parallel)
        self._FA4bbF = ddot(F, self._A4bbF, mode=(2, 2), parallel=self.parallel)

        return [
            self._hessian_uu(F, p, J),
            self._hessian_up(F, p, J),
            self._hessian_uJ(F, p, J),
            self._hessian_pp(F, p, J),
            self._hessian_pJ(F, p, J),
            self._hessian_JJ(F, p, J),
        ]

    def _hessian_uu(self, F, p=None, J=None):
        """Linearization w.r.t. displacements of variation of
        total potential energy w.r.t displacements.

        ..  code-block::

            Δ_u(δ_u(Π_int)) = ∫_V δF : (∂²ψ/(∂F∂F) + p ∂cof(F)/∂F) : ΔF dV

        Arguments
        ---------
        F : ndarray
            Deformation gradient
        p : ndarray
            Hydrostatic pressure
        J : ndarray
            Volume ratio

        Returns
        -------
        ndarray
            u,u - part of hessian
        """

        PbbA4bbF = self._Pbb + self._A4bbF
        PbbFA4bb = self._Pbb + self._FA4bb

        pJ9 = p * self._detF + self._PbbF / 9
        pJ3 = p * self._detF - self._PbbF / 3

        A4 = (
            self._A4bb
            + self._FA4bbF * dya(self._iFT, self._iFT, parallel=self.parallel) / 9
            - (
                dya(PbbA4bbF, self._iFT, parallel=self.parallel)
                + dya(self._iFT, PbbFA4bb, parallel=self.parallel)
            )
            / 3
            + pJ9 * dya(self._iFT, self._iFT, parallel=self.parallel)
            - pJ3 * cdya_il(self._iFT, self._iFT, parallel=self.parallel)
        )

        return A4

    def _hessian_pp(self, F, p, J):
        """Linearization w.r.t. pressure of variation of
        total potential energy w.r.t pressure.

        ..  code-block::

            Δ_p(δ_p(Π_int)) = ∫_V δp 0 Δp dV

        Arguments
        ---------
        F : ndarray
            Deformation gradient
        p : ndarray
            Hydrostatic pressure
        J : ndarray
            Volume ratio

        Returns
        -------
        ndarray
            p,p - part of hessian
        """
        return np.zeros_like(p)

    def _hessian_JJ(self, F, p, J):
        """Linearization w.r.t. volume ratio of variation of
        total potential energy w.r.t volume ratio.

        ..  code-block::

            Δ_J(δ_J(Π_int)) = ∫_V δJ ∂²ψ/(∂J∂J) ΔJ dV

        Arguments
        ---------
        F : ndarray
            Deformation gradient
        p : ndarray
            Hydrostatic pressure
        J : ndarray
            Volume ratio

        Returns
        -------
        ndarray
            J,J - part of hessian
        """

        return (self._FA4bbF - 2 * self._PbbF) / (9 * J**2)

    def _hessian_up(self, F, p, J):
        """Linearization w.r.t. pressure of variation of
        total potential energy w.r.t displacements.

        ..  code-block::

            Δ_p(δ_u(Π_int)) = ∫_V δF : J cof(F) Δp dV

        Arguments
        ---------
        F : ndarray
            Deformation gradient
        p : ndarray
            Hydrostatic pressure
        J : ndarray
            Volume ratio

        Returns
        -------
        ndarray
            u,p - part of hessian
        """

        return self._detF * self._iFT

    def _hessian_uJ(self, F, p, J):
        """Linearization w.r.t. volume ratio of variation of
        total potential energy w.r.t displacements.

        ..  code-block::

            Δ_J(δ_u(Π_int)) = ∫_V δF :  ∂²ψ/(∂F∂J) ΔJ dV

        Arguments
        ---------
        F : ndarray
            Deformation gradient
        p : ndarray
            Hydrostatic pressure
        J : ndarray
            Volume ratio

        Returns
        -------
        ndarray
            u,J - part of hessian
        """

        Ps = self._gradient_u(F, 0 * p, J)
        return (-self._FA4bbF / 3 * self._iFT + Ps + self._FA4bb) / (3 * J)

    def _hessian_pJ(self, F, p, J):
        """Linearization w.r.t. volume ratio of variation of
        total potential energy w.r.t pressure.

        ..  code-block::

            Δ_J(δ_p(Π_int)) = ∫_V δp (-1) ΔJ dV

        Arguments
        ---------
        F : ndarray
            Deformation gradient
        p : ndarray
            Hydrostatic pressure
        J : ndarray
            Volume ratio

        Returns
        -------
        ndarray
            p,J - part of hessian
        """
        return -np.ones_like(J)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from copy import deepcopy

import numpy as np


class DiscreteGeometry:
    """A discrete geometry with points, cells and optional a specified cell type.

    Parameters
    ----------
    points : ndarray
        Point coordinates.
    cells : ndarray
        Point-connectivity of cells.
    cell_type : str or None, optional
        An optional string in VTK-convention that specifies the cell type (default is
        None). Necessary when a discrete geometry is saved to a file.

    Attributes
    ----------
    points : ndarray
        Point coordinates.
    cells : ndarray
        Point-connectivity of cells.
    cell_type : str or None
        A string in VTK-convention that specifies the cell type.
    npoints : int
        Amount of points.
    dim : int
        Dimension of mesh point coordinates.
    ndof : int
        Amount of degrees of freedom.
    ncells : int
        Amount of cells.
    points_with_cells : array
        Array with points connected to cells.
    points_without_cells : array
        Array with points not connected to cells.
    cells_per_point : array
        Array which counts connected cells per point. Used for averaging results.

    See Also
    --------
    felupe.Mesh : A mesh with points, cells and optional a specified cell type.

    """

    def __init__(self, points, cells, cell_type=None):
        self.points = np.array(points)
        self.cells = np.array(cells)
        self.cell_type = cell_type

        self.update()

    def copy(self, points=None, cells=None, cell_type=None):
        """Return a deepcopy."""

        out = deepcopy(self)

        if points is not None or cells is not None or cell_type is not None:
            out.update(points=points, cells=cells, cell_type=cell_type)

        return out

    def update(self, points=None, cells=None, cell_type=None, callback=None):
        """Update the mesh with given points and cells arrays inplace. Optionally, a
        callback is evaluated.

        Parameters
        ----------
        points : ndarray or None
            New point coordinates (default is None). If None, it is unchanged.
        cells : ndarray
            New point-connectivity of cells (default is None). If None, it is unchanged.
        cell_type : str or None, optional
            New string in VTK-convention that specifies the cell type (default is
            None). If None, it is unchanged.
        callback : callable or None, optional
            A callable which is called after the mesh is updated (default is None).

        Examples
        --------
        ..  warning::
            If the points of a mesh are modified and a region was already created with
            the mesh, it is important to re-evaluate (reload) the
            :class:`~felupe.Region`.

        >>> import felupe as fem
        >>>
        >>> mesh = fem.Cube(n=6)
        >>> region = fem.RegionHexahedron(mesh)
        >>> field = fem.FieldContainer([fem.Field(region, dim=3)])
        >>>
        >>> new_points = mesh.rotate(angle_deg=-90, axis=2).points
        >>> mesh.update(points=new_points, callback=region.reload)

        See Also
        --------
        felupe.Region.reload : Reload the numeric region.
        """

        if points is not None:
            self.points = points

        if cells is not None:
            self.cells = cells

        if cell_type is not None:
            self.cell_type = cell_type

        # obtain dimensions
        self.npoints, self.dim = self.points.shape
        self.ndof = self.points.size
        self.ncells = self.cells.shape[0]

        # get number of cells per point
        points_in_cell, self.cells_per_point = np.unique(self.cells, return_counts=True)

        # check if there are points without cells
        if self.npoints != len(self.cells_per_point):
            self.point_has_cell = np.isin(np.arange(self.npoints), points_in_cell)
            # update "cells_per_point" ... cells per point
            cells_per_point = -np.ones(self.npoints, dtype=int)
            cells_per_point[points_in_cell] = self.cells_per_point
            self.cells_per_point = cells_per_point

            self.points_without_cells = np.arange(self.npoints)[~self.point_has_cell]
            self.points_with_cells = np.arange(self.npoints)[self.point_has_cell]
        else:
            self.points_without_cells = np.array([], dtype=int)
            self.points_with_cells = np.arange(self.npoints)

        if callable(callback):
            callback(self)

    @property
    def x(self):
        "Return the first column (x-component) of the points array."
        return self.points[:, 0]

    @property
    def y(self):
        "Return the second column (y-component) of the points array."
        return self.points[:, 1]

    @property
    def z(self):
        "Return the third column (z-component) of the points array."
        return self.points[:, 2]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from string import ascii_lowercase

import numpy as np

from ._scheme import Scheme


class GaussLegendre(Scheme):
    r"""An arbitrary-`order` Gauss-Legendre quadrature rule of dimension 1, 2 or 3 on
    the interval :math:`[-1, 1]`.

    Parameters
    ----------
    order : int
        The number of sample points :math:`n` minus one. The quadrature rule integrates
        degree :math:`2n-1` polynomials exactly.
    dim : int
        The dimension of the quadrature region.
    permute : bool, optional
        Permute the quadrature points according to the cell point orderings (default is
        True). This is supported for two and three dimensions as well as first and
        second order schemes. Otherwise this flag is silently ignored.

    Notes
    -----

    The approximation is given by

    ..  math::

        \int_{-1}^1 f(x) dx \approx \sum f(x_q) w_q

    with quadrature points :math:`x_q` and corresponding weights :math:`w_q`.

    Examples
    --------
    >>> import felupe as fem

    >>> fem.GaussLegendre(order=2, dim=3).screenshot()

    .. image:: images/quadrature.png

    """

    def __init__(self, order: int, dim: int, permute: bool = True):
        if dim not in [1, 2, 3]:
            raise ValueError("Wrong dimension.")

        x, w = np.polynomial.legendre.leggauss(1 + order)

        points = (
            np.stack(np.meshgrid(*([x] * dim), indexing="ij"))[::-1].reshape(dim, -1).T
        )

        idx = list(ascii_lowercase)[:dim]
        weights = np.einsum(", ".join(idx), *([w] * dim)).ravel()

        if permute and order == 1 and dim == 2:
            points = points[[0, 1, 3, 2]]
            weights = weights[[0, 1, 3, 2]]

        if permute and order == 1 and dim == 3:
            points = points[[0, 1, 3, 2, 4, 5, 7, 6]]
            weights = weights[[0, 1, 3, 2, 4, 5, 7, 6]]

        if permute and order == 2 and dim == 3:
            vertices = np.array([0, 2, 8, 6, 18, 20, 26, 24])
            edges = np.array([1, 5, 7, 3, 19, 23, 25, 21, 9, 11, 17, 15])
            faces = np.array([12, 14, 10, 16, 4, 22])
            volume = np.array([13])

            permute = np.concatenate((vertices, edges, faces, volume))

            points = points[permute]
            weights = weights[permute]

        super().__init__(points, weights)

    def inv(self):
        "Return the inverse quadrature scheme."

        points = self.points.copy()
        points[self.points != 0] = 1 / points[self.points != 0]

        return Scheme(points, self.weights)


class GaussLegendreBoundary(GaussLegendre):
    r"""An arbitrary-`order` Gauss-Legendre quadrature rule of `dim` 1, 2 or 3 on the
    interval ``[-1, 1]``.

    Parameters
    ----------
    order : int
        The number of sample points :math:`n` minus one. The quadrature rule integrates
        degree :math:`2n-1` polynomials exactly.
    dim : int
        The dimension of the quadrature region.
    permute : bool, optional
        Permute the quadrature points according to the cell point orderings (default is
        True).

    Notes
    -----

    The approximation is given by

    ..  math::

        \int_{-1}^1 f(x) dx \approx \sum f(x_q) w_q

    with quadrature points :math:`x_q` and corresponding weights :math:`w_q`.

    Examples
    --------
    >>> import felupe as fem

    >>> fem.GaussLegendreBoundary(order=2, dim=3).screenshot()

    .. image:: images/quadrature_boundary.png

    """

    def __init__(self, order: int, dim: int, permute: bool = True):
        super().__init__(order=order, dim=dim - 1, permute=permute)

        # reset dimension
        self.dim = dim

        if self.dim == 2 or self.dim == 3:
            # quadrature points projected onto first edge of a quad
            #                          or onto first face of a hexahedron
            self.points = np.hstack((self.points, -np.ones((len(self.points), 1))))

        else:
            raise ValueError("Given dimension not implemented (must be 2 or 3).")



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ._base import Element


class Triangle(Element):
    r"""A 2D triangle element formulation with linear shape functions.

    Notes
    -----
    The triangle element is defined by three points (0-2). [1]

    The shape functions :math:`\boldsymbol{h}` are given in terms of the coordinates
    :math:`(r,s)` [2]_.

    .. math::

       \boldsymbol{h}(r,s) = \begin{bmatrix}
               1-r-s \\
               r \\
               s
           \end{bmatrix}

    Examples
    --------
    >>> import felupe as fem

    >>> img = fem.Triangle().screenshot()

    ..  image:: images/triangle.png

    References
    ----------
    .. [1] W. Schroeder, K. Martin and B. Lorensen. The Visualization
       Toolkit, 4th ed. Kitware, 2006. ISBN: 978-1-930934-19-1.
    .. [2] K. J. Bathe, Finite element procedures, 2nd ed. K. J. Bathe, Watertown, MA,
       2014.
    """

    def __init__(self):
        super().__init__(shape=(3, 2))
        self.points = np.array([[0, 0], [1, 0], [0, 1]], dtype=float)
        self.cells = np.arange(len(self.points)).reshape(1, -1)
        self.cell_type = "triangle"

    def function(self, rs):
        "Return the shape functions at given coordinates (r, s)."
        r, s = rs
        return np.array([1 - r - s, r, s])

    def gradient(self, rs):
        "Return the gradient of shape functions at given coordinates (r, s)."
        r, s = rs
        return np.array([[-1, -1], [1, 0], [0, 1]], dtype=float)


class TriangleMINI(Element):
    r"""A 2D triangle element formulation with bubble-enriched linear shape functions.

    Notes
    -----
    The MINI triangle element is defined by four points (0-3). [1]

    The shape functions :math:`\boldsymbol{h}` are given in terms of the coordinates
    :math:`(r,s)`.

    .. math::

       \boldsymbol{h}(r,s) = \begin{bmatrix}
               1-r-s \\
               r \\
               s \\
               r s (1-r-s)
           \end{bmatrix}

    Examples
    --------
    >>> import felupe as fem

    >>> img = fem.TriangleMINI().screenshot(filename="triangle-mini.png")

    ..  image:: images/triangle-mini.png

    References
    ----------
    .. [1] W. Schroeder, K. Martin and B. Lorensen. The Visualization
       Toolkit, 4th ed. Kitware, 2006. ISBN: 978-1-930934-19-1.
    """

    def __init__(self, bubble_multiplier=1.0):
        super().__init__(shape=(4, 2))
        self.points = np.array([[0, 0], [1, 0], [0, 1], [1 / 3, 1 / 3]], dtype=float)
        self.cells = np.arange(len(self.points) - 1).reshape(1, -1)
        self.cell_type = "triangle"
        self.bubble_multiplier = bubble_multiplier

    def function(self, rs):
        "Return the shape functions at given coordinates (r, s)."
        r, s = rs
        a = self.bubble_multiplier
        return np.array([1 - r - s, r, s, a * r * s * (1 - r - s)])

    def gradient(self, rs):
        "Return the gradient of shape functions at given coordinates (r, s)."
        r, s = rs
        a = self.bubble_multiplier
        return np.array(
            [
                [-1, -1],
                [1, 0],
                [0, 1],
                [a * (s * (1 - r - s) - r * s), a * (r * (1 - r - s) - r * s)],
            ],
            dtype=float,
        )


class QuadraticTriangle(Element):
    r"""A 2D triangle element formulation with quadratic shape functions.

    Notes
    -----
    The quadratic triangle element is defined by six points (0-5). The element includes
    three mid-edge points besides the three triangle vertices. The ordering of the three
    points defining the element is point ids (0-2,3-5) where id #3 is the mid-edge point
    between points (0,1); id #4 is the mid-edge point between points (1,2); and id #5 is
    the mid-edge point between points (2,0). [1]

    The shape functions :math:`\boldsymbol{h}` are given in terms of the coordinates
    :math:`(r,s)` [2]_.

    .. math::

       \boldsymbol{h}(r,s) = \begin{bmatrix}
               1-r-s \\
               r \\
               s \\
               4 r (1-r-s) \\
               4 r s \\
               4 s (1-r-s)
           \end{bmatrix}

    Examples
    --------
    >>> import felupe as fem

    >>> img = fem.QuadraticTriangle().screenshot()

    ..  image:: images/triangle6.png

    References
    ----------
    .. [1] W. Schroeder, K. Martin and B. Lorensen. The Visualization
       Toolkit, 4th ed. Kitware, 2006. ISBN: 978-1-930934-19-1.
    .. [2] K. J. Bathe, Finite element procedures, 2nd ed. K. J. Bathe, Watertown, MA,
       2014.
    """

    def __init__(self):
        super().__init__(shape=(6, 2))
        self.points = np.zeros(self.shape)
        self.points[:3] = np.array([[0, 0], [1, 0], [0, 1]], dtype=float)
        self.points[3] = np.mean(self.points[[0, 1]], axis=0)
        self.points[4] = np.mean(self.points[[1, 2]], axis=0)
        self.points[5] = np.mean(self.points[[2, 0]], axis=0)

        self.cells = np.arange(len(self.points)).reshape(1, -1)
        self.cell_type = "triangle6"

    def function(self, rs):
        "Return the shape functions at given coordinates (r, s)."
        r, s = rs
        h = np.array(
            [1 - r - s, r, s, 4 * r * (1 - r - s), 4 * r * s, 4 * s * (1 - r - s)]
        )
        h[0] += -h[3] / 2 - h[5] / 2
        h[1] += -h[3] / 2 - h[4] / 2
        h[2] += -h[4] / 2 - h[5] / 2

        return h

    def gradient(self, rs):
        "Return the gradient of shape functions at given coordinates (r, s)."
        r, s = rs

        t1 = 1 - r - s
        t2 = r
        t3 = s

        dhdr_a = np.array([[-1, -1], [1, 0], [0, 1]], dtype=float)
        dhdr_b = np.array(
            [
                [4 * (t1 - t2), -4 * t2],
                [4 * t3, 4 * t2],
                [-4 * t3, 4 * (t1 - t2)],
            ]
        )
        dhdr = np.vstack((dhdr_a, dhdr_b))
        dhdr[0] += -dhdr[3] / 2 - dhdr[5] / 2
        dhdr[1] += -dhdr[3] / 2 - dhdr[4] / 2
        dhdr[2] += -dhdr[4] / 2 - dhdr[5] / 2

        return dhdr



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______ 
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___ 
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___ 
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""

import pytest

import felupe as fem
from felupe.math import ddot, dot, dya


def lform(v, w):
    return dot(w, v)


def lform_grad(v, F):
    return ddot(F, v)


def lformu(v, F, p):
    return ddot(F, v)


def lformp(q, F, p):
    return dot(p, q)


def bform(v, u, w):
    return dot(w, v) * dot(w, u)


def bform_grad(v, u, F):
    return ddot(F, v) * ddot(F, u)


def a_uu(v, u, F, p):
    return ddot(u, ddot(dya(F, F), v))


def a_up(v, r, F, p):
    return dot(p, r) * ddot(F, v)


def a_pp(q, r, F, p):
    return dot(p, q) * dot(r, p)


def pre(dim):
    mesh = fem.Cube(n=3)
    region = fem.RegionHexahedron(mesh)
    region0 = fem.RegionConstantHexahedron(mesh.dual(points_per_cell=1))
    u = fem.Field(region, dim=3)
    p = fem.Field(region0)
    up = fem.FieldContainer([u, p])

    return up


def test_form_decorator():
    field = pre(dim=3)
    F, p = field.extract()

    @fem.Form(v=field, u=field, grad_v=(True, False), grad_u=(True, False))
    def a():
        return (a_uu, a_up, a_pp)

    a.assemble(field, field, args=(F, p))

    @fem.Form(v=field, u=field, grad_v=None, grad_u=None)
    def a():
        return (a_uu, a_up, a_pp)

    a.assemble(field, field, args=(F, p))

    @fem.Form(v=field, grad_v=(True, False))
    def L():
        return (lformu, lformp)

    L.assemble(field, args=(F, p), parallel=False)
    L.assemble(field, args=(F, p), parallel=True)
    L.assemble(field, args=(F, p), parallel=False, sym=True)
    L.assemble(field, args=(F, p), parallel=True, sym=True)

    @fem.Form(v=field, grad_v=None)
    def L():
        return (lformu, lformp)

    L.integrate(field, args=(F, p), parallel=False)
    L.integrate(field, args=(F, p), parallel=True)
    L.integrate(field, args=(F, p), parallel=False, sym=True)
    L.integrate(field, args=(F, p), parallel=True, sym=True)

    L.assemble(field, args=(F, p), parallel=False)
    L.assemble(field, args=(F, p), parallel=True)
    L.assemble(field, args=(F, p), parallel=False, sym=True)
    L.assemble(field, args=(F, p), parallel=True, sym=True)


def test_linear_elastic():
    mesh = fem.Cube(n=11)
    region = fem.RegionHexahedron(mesh)
    displacement = fem.Field(region, dim=3)
    field = fem.FieldContainer([displacement])

    from felupe.math import ddot, sym, trace

    @fem.Form(
        v=field, u=field, grad_v=[True], grad_u=[True], kwargs={"mu": 1.0, "lmbda": 2.0}
    )
    def bilinearform():
        "A container for a bilinear form."

        def linear_elasticity(gradv, gradu, mu, lmbda):
            "Linear elasticity."

            de, e = sym(gradv), sym(gradu)
            return 2 * mu * ddot(de, e) + lmbda * trace(de) * trace(e)

        return [
            linear_elasticity,
        ]

    bilinearform.integrate(v=field, u=field, parallel=False)
    bilinearform.integrate(v=field, u=field, parallel=True)
    bilinearform.integrate(v=field, u=field, parallel=False, sym=True)
    bilinearform.integrate(v=field, u=field, parallel=True, sym=True)

    bilinearform.assemble(v=field, u=field, parallel=False)
    bilinearform.assemble(v=field, u=field, parallel=True)
    bilinearform.assemble(v=field, u=field, parallel=False, sym=True)
    bilinearform.assemble(v=field, u=field, parallel=True, sym=True)


if __name__ == "__main__":
    test_form_decorator()
    test_linear_elastic()



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ._view import ViewMaterial, ViewMaterialIncompressible


class ConstitutiveMaterial:
    "Base class for constitutive materials."

    def view(self, incompressible=False, **kwargs):
        """Create views on normal force per undeformed area vs. stretch curves for the
        elementary homogeneous deformations uniaxial tension/compression, planar shear
        and biaxial tension of a given isotropic material formulation.

        Parameters
        ----------
        incompressible : bool, optional
            A flag to enforce views on incompressible deformations (default is False).
        **kwargs : dict, optional
            Optional keyword-arguments for :class:`~felupe.ViewMaterial` or
            :class:`~felupe.ViewMaterialIncompressible`.

        Returns
        -------
        felupe.ViewMaterial or felupe.ViewMaterialIncompressible

        See Also
        --------
        felupe.ViewMaterial : Create views on normal force per undeformed area vs.
            stretch curves for the elementary homogeneous deformations uniaxial
            tension/compression, planar shear and biaxial tension of a given isotropic
            material formulation.
        felupe.ViewMaterialIncompressible : Create views on normal force per undeformed
            area vs. stretch curves for the elementary homogeneous incompressible
            deformations uniaxial tension/compression, planar shear and biaxial tension
            of a given isotropic material formulation.
        """

        View = ViewMaterial
        if incompressible:
            View = ViewMaterialIncompressible

        return View(self, **kwargs)

    def plot(self, incompressible=False, **kwargs):
        """Return a plot with normal force per undeformed area vs. stretch curves for
        the elementary homogeneous deformations uniaxial tension/compression, planar
        shear and biaxial tension of a given isotropic material formulation.

        Parameters
        ----------
        incompressible : bool, optional
            A flag to enforce views on incompressible deformations (default is False).
        **kwargs : dict, optional
            Optional keyword-arguments for :class:`~felupe.ViewMaterial` or
            :class:`~felupe.ViewMaterialIncompressible`.

        Returns
        -------
        matplotlib.axes.Axes

        See Also
        --------
        felupe.ViewMaterial : Create views on normal force per undeformed area vs.
            stretch curves for the elementary homogeneous deformations uniaxial
            tension/compression, planar shear and biaxial tension of a given isotropic
            material formulation.
        felupe.ViewMaterialIncompressible : Create views on normal force per undeformed
            area vs. stretch curves for the elementary homogeneous incompressible
            deformations uniaxial tension/compression, planar shear and biaxial tension
            of a given isotropic material formulation.
        """

        return self.view(incompressible=incompressible, **kwargs).plot()

    def screenshot(self, filename="umat.png", incompressible=False, **kwargs):
        """Save a screenshot with normal force per undeformed area vs. stretch curves
        for the elementary homogeneous deformations uniaxial tension/compression, planar
        shear and biaxial tension of a given isotropic material formulation.

        Parameters
        ----------
        filename : str, optional
            The filename of the screenshot (default is "umat.png").
        incompressible : bool, optional
            A flag to enforce views on incompressible deformations (default is False).
        **kwargs : dict, optional
            Optional keyword-arguments for :class:`~felupe.ViewMaterial` or
            :class:`~felupe.ViewMaterialIncompressible`.

        Returns
        -------
        matplotlib.axes.Axes

        See Also
        --------
        felupe.ViewMaterial : Create views on normal force per undeformed area vs.
            stretch curves for the elementary homogeneous deformations uniaxial
            tension/compression, planar shear and biaxial tension of a given isotropic
            material formulation.
        felupe.ViewMaterialIncompressible : Create views on normal force per undeformed
            area vs. stretch curves for the elementary homogeneous incompressible
            deformations uniaxial tension/compression, planar shear and biaxial tension
            of a given isotropic material formulation.
        """

        import matplotlib.pyplot as plt

        ax = self.plot(incompressible=incompressible, **kwargs)
        fig = ax.get_figure()
        fig.savefig(filename)
        plt.close(fig)

        return ax

    def __and__(self, other_material):
        return CompositeMaterial(self, other_material)


class CompositeMaterial(ConstitutiveMaterial):
    """A composite material with two constitutive materials merged. State variables are
    only considered for the first material.

    Parameters
    ----------
    material : ConstitutiveMaterial
        First constitutive material.
    other_material : ConstitutiveMaterial
        Second constitutive material.

    Notes
    -----
    ..  warning::
        Do not merge two constitutive materials with the same keys of material
        parameters. In this case, the values of these material parameters are taken from
        the first constitutive material.

    Examples
    --------
    >>> import felupe as fem
    >>>
    >>> nh = fem.NeoHooke(mu=1.0)
    >>> vol = fem.Volumetric(bulk=2.0)
    >>> umat = nh & vol
    >>> ax = umat.plot()

    ..  image:: images/umat_composite.png
        :width: 400px

    """

    def __init__(self, material, other_material):
        self.materials = [material, other_material]
        self.kwargs = {**other_material.kwargs, **material.kwargs}
        self.x = material.x

    def gradient(self, x, **kwargs):
        gradients = [material.gradient(x, **kwargs) for material in self.materials]
        nfields = len(x) - 1
        P = [np.sum([grad[i] for grad in gradients], axis=0) for i in range(nfields)]
        statevars_new = gradients[0][-1]
        return [*P, statevars_new]

    def hessian(self, x, **kwargs):
        hessians = [material.hessian(x, **kwargs) for material in self.materials]
        nfields = len(x) - 1
        return [np.sum([hess[i] for hess in hessians], axis=0) for i in range(nfields)]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
import os

import numpy as np

from ..math import dot, eigh, eigvalsh, tovoigt, transpose
from ..tools._misc import logo, runs_on


def displacement(field, substep=None):
    "Displacement Vector"
    u = field[0].values
    return np.pad(u, ((0, 0), (0, 3 - u.shape[1])))


def deformation_gradient(field, substep=None):
    "Deformation Gradient"
    F = field[0].extract()
    return [F.mean(-2).transpose([2, 0, 1])]


def log_strain_principal(field, substep=None):
    "Principal Values of Logarithmic Strain"
    u = field[0]
    F = u.extract()
    stretch = np.sqrt(eigvalsh(dot(transpose(F), F)))[::-1]
    strain = np.log(stretch).mean(-2)
    return [strain.T]


def log_strain(field, substep=None):
    "Lagrangian Logarithmic Strain Tensor"
    u = field[0]
    F = u.extract()
    w, v = eigh(dot(transpose(F), F))
    stretch = np.sqrt(w)
    strain = np.einsum("a...,ai...,aj...->ij...", np.log(stretch), v, v)
    return [tovoigt(strain.mean(-2), True).T]


def print_header():
    print("\n".join([logo(), runs_on(), "", "Run Job", "======="]))


class Job:
    r"""A job with a list of steps and a method to evaluate them.

    Parameters
    ----------
    steps : list of Step
        A list with steps, where each step subsequently depends on the solution of the
        previous step.
    callback : callable, optional
        A callable which is called after each completed substep. Function signature must
        be ``lambda stepnumber, substepnumber, substep, **kwargs: None``, where
        ``substep`` is an instance of :class:`~felupe.tools.NewtonResult`. The field
        container of the completed substep is available as ``substep.x``. Default
        is ``callback=lambda stepnumber, substepnumber, substep, **kwargs: None``.
    **kwargs : dict, optional
        Optional keyword-arguments for the ``callback`` function.

    Attributes
    ----------
    steps : list of Step
        A list with steps, where each step subsequently depends on the solution of the
        previous step.
    nsteps : int
        The number of steps.
    callback : callable
        A callable which is called after each completed substep. Function signature must
        be ``lambda stepnumber, substepnumber, substep: None``, where ``substep`` is an
        instance of :class:`~felupe.tools.NewtonResult`. THe field container of the
        completed substep is available as ``substep.x``.
    timetrack : list of int
        A list with times at which the results are written to the XDMF result file.
    fnorms : list of list of float
        List with norms of the objective function for each completed substep of each
        step. See also class:`~felupe.tools.NewtonResult`.
    kwargs : dict
        Optional keyword-arguments for the ``callback`` function.

    Examples
    --------
    >>> import felupe as fem
    >>>
    >>> mesh = fem.Cube(n=6)
    >>> region = fem.RegionHexahedron(mesh)
    >>> field = fem.FieldContainer([fem.Field(region, dim=3)])
    >>>
    >>> boundaries = fem.dof.symmetry(field[0])
    >>> boundaries["clamped"] = fem.Boundary(field[0], fx=1, skip=(True, False, False))
    >>> boundaries["move"] = fem.Boundary(field[0], fx=1, skip=(False, True, True))
    >>>
    >>> umat = fem.NeoHooke(mu=1, bulk=2)
    >>> solid = fem.SolidBody(umat, field)
    >>>
    >>> move = fem.math.linsteps([0, 1], num=5)
    >>> step = fem.Step(items=[solid], ramp={boundaries["move"]: move}, boundaries=boundaries)
    >>>
    >>> job = fem.Job(steps=[step]).evaluate()
    >>> ax = solid.imshow("Principal Values of Cauchy Stress")

    See Also
    --------
    Step : A Step with multiple substeps, subsequently depending on the solution
        of the previous substep.
    CharacteristicCurve : A job with a list of steps and a method to evaluate them.
        Force-displacement curve data is tracked during evaluation for a given
        :class:`~felupe.Boundary`.
    tools.NewtonResult : A data class which represents the result found by
        Newton's method.

    """

    def __init__(
        self,
        steps,
        callback=lambda stepnumber, substepnumber, substep, **kwargs: None,
        **kwargs,
    ):
        self.steps = steps
        self.nsteps = len(steps)
        self.callback = callback
        self.timetrack = []
        self.fnorms = []
        self.kwargs = kwargs

    def _write(self, writer, time, substep, point_data, cell_data):
        field = substep.x
        kwargs = dict(field=field, substep=substep)
        writer.write_data(
            time,
            point_data={key: value(**kwargs) for key, value in point_data.items()},
            cell_data={key: value(**kwargs) for key, value in cell_data.items()},
        )

    def evaluate(
        self,
        filename=None,
        mesh=None,
        point_data=None,
        cell_data=None,
        point_data_default=True,
        cell_data_default=True,
        verbose=True,
        parallel=False,
        **kwargs,
    ):
        """Evaluate the steps.

        Parameters
        ----------
        filename : str or None, optional
            The filename of the XDMF result file. Must include the file extension
            ``my_result.xdmf``. If None, no result file is writte during evaluation.
            Default is None.
        mesh : Mesh or None, optional
            A mesh which is used for the XDMF time series writer. If None, it is taken
            from the field of the first item of the first step if no keyword argument
            ``x0`` is given. If None and ``x0=field``, the mesh is taken from the ``x0``
            field container. Default is None.
        point_data : dict or None, optional
            Additional dict of point-data for the meshio XDMF time series writer.
        cell_data : dict or None, optional
            Additional dict of cell-data for the meshio XDMF time series writer.
        point_data_default : bool, optional
            Flag to write default point-data to the XDMF result file. This includes
            ``"Displacement"``. Default is True.
        cell_data_default : bool, optional
            Flag to write default cell-data to the XDMF result file. This includes
            ``"Principal Values of Logarithmic Strain"``, ``"Logarithmic Strain"`` and
            ``"Deformation Gradient"``. Default is True.
        verbose : bool or int, optional
            Verbosity level to control how messages are printed during evaluation. If
            1 or True and ``tqdm`` is installed, a progress bar is shown. If ``tqdm`` is
            missing or verbose is 2, more detailed text-based messages are printed.
            Default is True. If the environmental variable FELUPE_VERBOSE is set and
            its value is ``false``, then this argument is ignored and logging is turned
            off.
        parallel : bool, optional
            Flag to use a threaded version of :func:`numpy.einsum` during assembly.
            Requires ``einsumt``. This may add additional overhead to small-sized
            problems. Default is False.
        **kwargs : dict
            Optional keyword arguments for :meth:`~felupe.Step.generate`. If
            ``parallel=True``, it is added as ``kwargs["parallel"] = True`` to the dict
            of additional keyword arguments. If ``x0`` is present in ``kwargs.keys()``,
            it is used as the mesh for the XDMF time series writer.

        Returns
        -------
        Job
            The job object.

        Notes
        -----
        Requires ``meshio`` and ``h5py`` if ``filename`` is not None. Also requires
        ``tqdm`` for an interactive progress bar if ``verbose=True``.

        See Also
        --------
        Step : A Step with multiple substeps, subsequently depending on the solution
            of the previous substep.
        CharacteristicCurve : A job with a list of steps and a method to evaluate them.
            Force-displacement curve data is tracked during evaluation for a given
            :class:`~felupe.Boundary` condition.
        tools.NewtonResult : A data class which represents the result found by
            Newton's method.
        """
        VERBOSE = os.environ.get("FELUPE_VERBOSE")
        if VERBOSE is None:
            verbose = verbose
        else:
            verbose = VERBOSE == "true"

        if verbose:
            try:
                from tqdm import tqdm
            except ModuleNotFoundError:
                verbose = 2

        if verbose == 2:
            print_header()

        if parallel:
            if "kwargs" not in kwargs.keys():
                kwargs["kwargs"] = {}
            kwargs["kwargs"]["parallel"] = True

        time = 0

        if filename is not None:
            from meshio.xdmf import TimeSeriesWriter

            if mesh is None:
                if "x0" in kwargs.keys():
                    mesh = kwargs["x0"].region.mesh.as_meshio()
                else:
                    mesh = self.steps[0].items[0].field.region.mesh.as_meshio()

            pdata = point_data_default if point_data_default is True else {}
            cdata = cell_data_default if cell_data_default is not None else {}

            pdata = {}
            cdata = {}

            if point_data_default is True:
                pdata = {"Displacement": displacement}

            if cell_data_default is True:
                cdata = {
                    "Principal Values of Logarithmic Strain": log_strain_principal,
                    "Logarithmic Strain": log_strain,
                    "Deformation Gradient": deformation_gradient,
                }

            if point_data is None:
                point_data = {}

            if cell_data is None:
                cell_data = {}

        else:  # fake a mesh and a TimeSeriesWriter
            from contextlib import nullcontext

            TimeSeriesWriter = nullcontext

        with TimeSeriesWriter(filename) as writer:
            if filename is not None:
                writer.write_points_cells(mesh.points, mesh.cells)

            if verbose == 1:
                total = sum([step.nsubsteps for step in self.steps])
                progress_bar = tqdm(total=total, unit="substep")

            for j, step in enumerate(self.steps):
                newton_verbose = False
                if verbose == 2:
                    print(f"Begin Evaluation of Step {j + 1}.")
                    newton_verbose = True

                substeps = step.generate(verbose=newton_verbose, **kwargs)
                for i, substep in enumerate(substeps):
                    self.fnorms.append(substep.fnorms)
                    if verbose == 2:
                        _substep = f"Substep {i + 1}/{step.nsubsteps}"
                        _step = f"Step {j + 1}/{self.nsteps}"

                        print(f"{_substep} of {_step} successful.")

                    self.callback(j, i, substep, **self.kwargs)

                    # update x0 after each completed substep
                    if "x0" in kwargs.keys():
                        kwargs["x0"].link(substep.x)

                    self.timetrack.append(time)

                    if filename is not None:
                        self._write(
                            writer=writer,
                            time=time,
                            substep=substep,
                            point_data={**pdata, **point_data},
                            cell_data={**cdata, **cell_data},
                        )

                    time += 1

                    if verbose == 1:
                        progress_bar.update(1)

            if verbose == 1:
                progress_bar.close()

        return self



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from scipy.interpolate import interp1d
from scipy.sparse import issparse

from ..math import cross


def force(field, forces, boundary):
    "Evaluate the force vector sum on points of a boundary."

    if issparse(forces):
        forces = forces.toarray()

    forces_first_field = np.split(forces, field.offsets)[0]
    dim = field[0].dim

    return ((forces_first_field.reshape(-1, dim))[boundary.points]).sum(axis=0)


def moment(field, forces, boundary, centerpoint=np.zeros(3)):
    "Evaluate the moment vector sum on points of a boundary at a given center point."

    if issparse(forces):
        forces = forces.toarray()

    dim = field[0].dim
    centerpoint = np.asarray(centerpoint).reshape(1, -1)[:, :dim]

    displacements = field[0].values
    force = (np.split(forces, field.offsets)[0]).reshape(-1, dim)

    moments = cross(
        (field.region.mesh.points + displacements - centerpoint)[boundary.points].T,
        force[boundary.points].T,
    ).T

    return moments.sum(axis=0)


def curve(x, y, num=50):
    "Interpolate a curve from given (x, y) data."

    kind = [None, "linear", "quadratic", "cubic"][min(len(y), 4) - 1]

    f = interp1d(x[: len(y)], y, kind=kind)

    xt = x[: len(y)]
    xx = np.linspace(xt[0], xt[-1], num=num)

    return np.array([xt, y]), np.array([xx, f(xx)])



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ..math import dya
from ._base import ConstitutiveMaterial


class OgdenRoxburgh(ConstitutiveMaterial):
    r"""`Ogden-Roxburgh <https://doi.org/10.1098%2Frspa.1999.0431>`_ Pseudo-Elastic
    material formulation for an isotropic treatment of the load-history dependent
    Mullins-softening of rubber-like materials.

    Parameters
    ----------
    material : NeoHooke, Hyperelastic, Material or MaterialAD
        An isotropic hyperelastic (user) material definition.
    r : float
        Reciprocal value of the maximum relative amount of softening. i.e. ``r=3`` means
        the shear modulus of the base material scales down from :math:`1` (no softening)
        to :math:`1 - 1/3 = 2/3` (maximum softening).
    m : float
        The initial Mullins softening modulus.
    beta : float
        Maximum deformation-dependent part of the Mullins softening modulus.

    Notes
    -----

    ..  math::

        \eta(\psi, \psi_{max}) &= 1 - \frac{1}{r} erf\left( \frac{\psi_{max} - \psi}
        {m + \beta~\psi_{max}} \right)

        \boldsymbol{P} &= \eta \frac{\partial \psi}{\partial \boldsymbol{F}}

        \mathbb{A} &= \frac{\partial^2 \psi}{\partial \boldsymbol{F} \partial
        \boldsymbol{F}} + \frac{\partial \eta}{\partial \psi} \frac{\partial \psi}
        {\partial \boldsymbol{F}} \otimes \frac{\partial \psi}{\partial \boldsymbol{F}}

    Examples
    --------
    >>> import felupe as fem
    >>>
    >>> neo_hooke = fem.NeoHooke(mu=1.0)
    >>> umat = fem.OgdenRoxburgh(material=neo_hooke, r=3.0, m=1.0, beta=0.0)
    >>>
    >>> ax = umat.plot(
    >>>     ux=fem.math.linsteps([1, 1.5, 1, 2, 1, 2.5, 1], num=15),
    >>>     ps=None,
    >>>     bx=None,
    >>>     incompressible=True,
    >>> )

    .. image:: images/umat.png
       :width: 400px

    """

    def __init__(self, material, r, m, beta):
        # isotropic hyperelastic material formulation
        self.material = self.fun = material

        # ogden-roxburgh material parameters
        self.r = r
        self.m = m
        self.beta = beta

        self.kwargs = {
            "r": self.r,
            "m": self.m,
            "beta": self.beta,
            **self.material.kwargs,
        }

        # initial variables for calling
        # ``self.gradient(self.x)`` and ``self.hessian(self.x)``
        self.x = [np.eye(3), np.zeros(1)]

    def gradient(self, x):
        # unpack variables into deformation gradient and state variables
        F, statevars = x[0], x[-1]

        # material parameters alias
        r, m, beta = self.r, self.m, self.beta

        # isotropic material formulation: evaluate
        # * the strain energy function and
        # * the first Piola-Kirchhoff stress tensor
        W = self.material.function([F, statevars])[0]
        P = self.material.gradient([F, statevars])[0]

        # get the maximum load-history strain energy function
        Wmax = np.maximum(W, statevars[0])
        z = (Wmax - W) / (m + beta * Wmax)

        # softening function
        eta = 1 - np.tanh(z) / r

        # update the state variables
        statevars_new = statevars.copy()
        statevars_new[0] = Wmax

        return [eta * P, statevars_new]

    def hessian(self, x):
        # unpack variables into deformation gradient and state variables
        F, statevars = x[0], x[-1]

        # material parameters alias
        r, m, beta = self.r, self.m, self.beta

        # isotropic material formulation: evaluate
        # * the strain energy function and
        # * the first Piola-Kirchhoff stress tensor as well as
        # * the according fourth-order elasticity tensor
        W = self.material.function([F, statevars])[0]
        P = self.material.gradient([F, statevars])[0]
        A = self.material.hessian([F, statevars])[0]

        # get the maximum load-history strain energy function
        Wmax = np.maximum(W, statevars[0])
        z = (Wmax - W) / (m + beta * Wmax)

        # softening function
        eta = 1 - np.tanh(z) / r

        # derivative of softening function
        detadz = (-1 / r) * 1 / np.cosh(z) ** 2
        dzdW = -1 / (m + beta * Wmax)
        detadW = detadz * dzdW

        # set non-softened derivative to zero
        detadW[np.isclose(eta, 1)] = 0

        return [eta * A + detadW * dya(P, P)]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from copy import deepcopy

import numpy as np

from ._mesh import Mesh
from ._tools import merge_duplicate_points as sweep
from ._tools import stack


class MeshContainer:
    """A container which operates on a list of meshes with identical dimensions.

    Parameters
    ----------
    meshes : list of Mesh
        A list of meshes which are organized by the mesh container.
    merge : bool, optional
        Flag to merge duplicate mesh points. This changes the cells arrays of the
        meshes. Default is False.
    decimals : float or None, optional
        Precision decimals for merging duplicated mesh points. Only relevant if
        merge=True. Default is None.

    Notes
    -----
    All meshes are modified to refer to the same points array. By default, the points
    arrays from the given list of meshes is concatenated and the cells arrays are
    modified accordingly. Optionally, the points array may be merged on duplicated
    points.

    Examples
    --------
    >>> import felupe as fem
    >>>
    >>> cube = fem.Cube(n=3)
    >>> cylinder = fem.Circle().expand(n=2)
    >>> mesh = fem.MeshContainer([cube, cylinder])
    >>> mesh
    <felupe mesh container object>
      Number of points: 61
      Number of cells:
        hexahedron: 8
        hexahedron: 12

    The cells array of the second mesh starts with an offset

    >>> mesh.meshes[1].cells.min()
    27

    identical to the number of points from the first mesh.

    >>> cube.npoints
    27

    If the container is created with ``merge=True``, then the number of points is lower
    than before.

    >>> mesh = fem.MeshContainer([cube, cylinder], merge=True)
    >>> mesh
    <felupe mesh container object>
      Number of points: 51
      Number of cells:
        hexahedron: 8
        hexahedron: 12

    ..  image:: images/container.png
        :width: 400px

    """

    def __init__(self, meshes, merge=False, decimals=None):
        # obtain the dimension from the first mesh
        self.dim = meshes[0].dim

        # init points and list of meshes
        self.points = np.zeros((0, self.dim))
        self.meshes = []

        # append all meshes
        [self.append(mesh) for mesh in meshes]

        if merge:
            self.merge_duplicate_points(decimals=decimals)

    def append(self, mesh):
        "Append a :class:`~felupe.Mesh` to the list of meshes."

        # number of points
        points = np.vstack([self.points, mesh.points])
        self.meshes.append(Mesh(points, mesh.cells + len(self.points), mesh.cell_type))

        # ensure identical points-arrays
        for i, m in enumerate(self.meshes):
            self.meshes[i].points = self.points = points

    def pop(self, index):
        "Pop an item of the list of meshes."
        item = self.meshes.pop(index)
        return item

    def cells(self):
        "Return a list of tuples with cell-types and cell-connectivities."
        return [(mesh.cell_type, mesh.cells) for mesh in self.meshes]

    def merge_duplicate_points(self, decimals=None):
        "Merge duplicate points and update the meshes."

        # sweep points
        for i, mesh in enumerate(self.meshes):
            self.meshes[i] = sweep(mesh, decimals=decimals)

        # ensure identical points-arrays
        points = self.meshes[0].points
        for i, m in enumerate(self.meshes):
            self.meshes[i].points = self.points = points

    def stack(self):
        """Stack cell-blocks with same cell-types into a single mesh.

        Returns
        -------
        Mesh
            The stacked mesh.

        Notes
        -----
        The ``points``-array is taken from the first mesh. The ``cells``-array of the
        stacked mesh is created by a vertical stack of the ``cells``-arrays.

        Examples
        --------
        Two quad meshes with identical point arrays should be stacked into a single mesh.

        >>> import felupe as fem

        >>> mesh = fem.Rectangle(n=11)
        >>> rect1, rect2 = mesh.copy(), mesh.copy()
        >>> rect1.update(cells=mesh.cells[: 40])
        >>> rect2.update(cells=mesh.cells[-50:])
        >>> container = fem.MeshContainer([rect1, rect2])

        >>> mesh = container.stack()
        >>> mesh
        <felupe Mesh object>
          Number of points: 121
          Number of cells:
            quad: 90

        ..  image:: images/mesh_stack.png
            :width: 400px

        See Also
        --------
        felupe.mesh.stack : Stack cell-blocks from meshes with identical points-array
            and cell-types.
        """
        if all(mesh.cell_type == self.meshes[0].cell_type for mesh in self.meshes):
            return stack(self.meshes)
        else:
            message = [
                "Meshes of the MeshContainer can't be stacked.",
                "List of meshes must have same cell-types.",
            ]
            raise TypeError(" ".join(message))

    def as_meshio(self, combined=True, **kwargs):
        "Export a (combined) mesh object as :class:`meshio.Mesh`."

        import meshio

        if not combined:
            cells = [
                meshio.CellBlock(cell_type, data) for cell_type, data in self.cells()
            ]

        else:
            cells = {}
            for mesh in self.meshes:
                if mesh.cell_type not in cells.keys():
                    cells[mesh.cell_type] = mesh.cells
                else:
                    cells[mesh.cell_type] = np.vstack(
                        [cells[mesh.cell_type], mesh.cells]
                    )

        return meshio.Mesh(self.points, cells, **kwargs)

    def copy(self):
        "Return a deepcopy of the mesh container."
        return deepcopy(self)

    def plot(self, *args, colors=None, **kwargs):
        """Plot the meshes of the mesh container.

        See Also
        --------
        felupe.Scene.plot: Plot method of a scene.
        """

        if colors is None:
            import matplotlib.colors as mcolors
            import pyvista

            colors = [
                pyvista.global_theme.color,
                *list(mcolors.TABLEAU_COLORS.values())[1:],
            ]

        plotter = None
        for mesh, color in zip(self.meshes, colors):
            plotter = mesh.view().plot(
                *args,
                show_undeformed=False,
                color=color,
                plotter=plotter,
                opacity=0.99,
                **kwargs,
            )

        return plotter

    def screenshot(
        self,
        *args,
        filename="mesh.png",
        transparent_background=None,
        scale=None,
        colors=None,
        **kwargs,
    ):
        """Take a screenshot of the meshes of the mesh container.

        See Also
        --------
        pyvista.Plotter.screenshot: Take a screenshot of a PyVista plotter.
        """

        if colors is None:
            import matplotlib.colors as mcolors
            import pyvista

            colors = [
                pyvista.global_theme.color,
                *list(mcolors.TABLEAU_COLORS.values())[1:],
            ]

        plotter = None
        for mesh, color in zip(self.meshes, colors):
            plotter = mesh.plot(
                *args,
                off_screen=True,
                color=color,
                plotter=plotter,
                opacity=0.99,
                **kwargs,
            )

        return plotter.screenshot(
            filename=filename,
            transparent_background=transparent_background,
            scale=scale,
        )

    def imshow(self, ax=None, *args, **kwargs):
        """Take a screenshot of the meshes of the mesh container, show the image data in
        a figure and return the ax.
        """

        if ax is None:
            import matplotlib.pyplot as plt

            fig, ax = plt.subplots()

        ax.imshow(self.screenshot(*args, filename=None, **kwargs))
        ax.set_axis_off()

        return ax

    def __iadd__(self, mesh):
        self.append(mesh)
        return self

    def __getitem__(self, index):
        return self.meshes[index]

    def __repr__(self):
        header = "<felupe mesh container object>"
        points = f"  Number of points: {len(self.points)}"
        cells_header = "  Number of cells:"
        cells = []

        for cells_type, cells_data in self.cells():
            cells.append(f"    {cells_type}: {len(cells_data)}")

        return "\n".join([header, points, cells_header, *cells])

    def __str__(self):
        return self.__repr__()



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ._helpers import mesh_or_data


@mesh_or_data
def dual(
    points,
    cells,
    cell_type,
    points_per_cell=None,
    disconnect=True,
    calc_points=False,
    offset=0,
    npoints=None,
):
    """Create a new dual mesh with given points per cell.

    Parameters
    ----------
    points : list or ndarray
        Original point coordinates.
    cells : list or ndarray
        Original point-connectivity of cells.
    cell_type : str
        A string in VTK-convention that specifies the cell type.
    points_per_cell : int or None, optional
        Number of points per cell, must be equal or lower than ``cells.shape[1]`` (
        default is None). If None, all points per cell are considered for the dual mesh.
    disconnect : bool, optional
        A flag to disconnect the mesh (each cell has its own points). Default is True.
    calc_points : bool, optional
        A flag to calculate the point coordinates for the dual mesh (default is False).
        If False, the points array is filled with zeros.
    offset : int, optional
        An offset to be added to the cells array (default is 0).
    npoints : int or None, optional
        Number of points for the dual mesh. If the given number of points is greater
        than ``npoints * points_per_cell``, then the missing points are added to the
        points array (filled with zeros). Default is None.

    Returns
    -------
    points : ndarray
        Modified point coordinates.
    cells : list or ndarray
        Modified point-connectivity of cells.
    cell_type : str or None
        A string in VTK-convention that specifies the cell type.

    Notes
    -----
    ..  note::
        The points array of the dual mesh always has a shape of
        ``(npoints * points_per_cell, dim)``.

    Examples
    --------
    >>> import felupe as fem
    >>>
    >>> mesh = fem.Rectangle(n=5).add_midpoints_edges()
    >>> region = fem.RegionQuadraticQuad(mesh=mesh)
    >>>
    >>> mesh_dual = fem.mesh.dual(mesh, points_per_cell=1, disconnect=False)
    >>> region_dual = fem.RegionConstantQuad(
    >>>     mesh_dual, quadrature=region.quadrature, grad=False
    >>> )
    >>>
    >>> displacement = fem.FieldPlaneStrain(region, dim=2)
    >>> pressure = fem.Field(region_dual)
    >>> field = fem.FieldContainer([displacement, pressure])

    See Also
    --------
    felupe.Mesh.dual : Create a new dual mesh with given points per cell.

    """

    ncells = len(cells)
    dim = points.shape[1]
    cell_type_new = None

    if points_per_cell is None:
        points_per_cell = cells.shape[1]
        cell_type_new = cell_type

    if disconnect:
        cells_new = np.arange(ncells * points_per_cell).reshape(ncells, points_per_cell)
    else:
        cells_new = cells[:, :points_per_cell]

    if calc_points:
        points_new = points[cells[:, :points_per_cell]].reshape(-1, dim)
    else:
        points_new = np.broadcast_to(
            np.zeros((1, dim), dtype=int), (ncells * points_per_cell, dim)
        )

    if offset > 0:
        cells_new += offset
        points_new = np.pad(points_new, ((offset, 0), (0, 0)))

    if npoints is not None and npoints > len(points_new):
        points_new = np.pad(points_new, ((npoints - len(points_new), 0), (0, 0)))
    elif npoints is None and cells_new.max() > len(points_new):
        points_new = np.pad(
            points_new, ((1 + cells_new.max() - len(points_new), 0), (0, 0))
        )

    return points_new, cells_new, cell_type_new



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from ._helpers import Assemble, Results


class FormItem:
    r"""An item to be used in a :class:`felupe.Step` with bilinear and optional linear
    form objects based on weak-forms with methods for integration and assembly of
    vectors / sparse matrices.

    Parameters
    ----------
    bilinearform : Form or None, optional
        A bilinear form object (default is None). If None, the resulting matrix will be
        filled with zeros.
    linearform : Form or None, optional
        A linear form object (default is None). If None, the resulting vector will be
        filled with zeros.
    sym : bool, optional
        Flag to active symmetric integration/assembly for bilinear forms (default is
        False).
    args : tuple or None, optional
        Tuple with initial optional weakform-arguments (default is None).
    kwargs : dict or None, optional
        Dictionary with initial optional weakform-keyword-arguments (default is None).

    Examples
    --------
    >>> import felupe as fem
    >>> from felupe.math import ddot, sym, trace

    >>> mesh = fem.Cube(n=11)
    >>> region = fem.RegionHexahedron(mesh)
    >>> field = fem.FieldContainer([fem.Field(region, dim=3)])
    >>> boundaries, loadcase = fem.dof.uniaxial(field, clamped=True)

    >>> @fem.Form(v=field, u=field, grad_v=[True], grad_u=[True])
    >>> def bilinearform():
    >>>     def a(gradv, gradu, μ=1.0, λ=2.0):
    >>>         δε, ε = sym(gradv), sym(gradu)
    >>>         return 2 * μ * ddot(δε, ε) + λ * trace(δε) * trace(ε)
    >>>     return [a]

    >>> item = fem.FormItem(bilinearform, linearform=None, sym=True)
    >>> step = fem.Step(items=[item], boundaries=boundaries)
    >>> fem.Job(steps=[step]).evaluate()

    See Also
    --------
    felupe.Form : A function decorator for a linear- or bilinear-form object.
    felupe.Step : A Step with multiple substeps.

    """

    def __init__(
        self, bilinearform=None, linearform=None, sym=False, args=None, kwargs=None
    ):
        self.bilinearform = bilinearform
        self.linearform = linearform

        self.sym = sym
        self.args = args
        self.kwargs = kwargs

        self.results = Results(stress=False, elasticity=False)
        self.assemble = Assemble(vector=self._vector, matrix=self._matrix)

        if self.bilinearform is not None:
            self.field = self.bilinearform.form.v.field
        else:
            if self.linearform is not None:
                self.field = self.linearform.form.v.field

    def _vector(self, field=None, parallel=False):
        if field is not None:
            self.field = field

        if self.linearform is not None:
            self.results.force = self.linearform.assemble(
                v=self.field, parallel=parallel, args=self.args, kwargs=self.kwargs
            )

        else:
            from scipy.sparse import csr_matrix

            self.results.force = csr_matrix((sum(self.field.fieldsizes), 1))

        return self.results.force

    def _matrix(self, field=None, parallel=False):
        if field is not None:
            self.field = field

        if self.bilinearform is not None:
            self.results.stiffness = self.bilinearform.assemble(
                v=self.field,
                u=self.field,
                parallel=parallel,
                sym=self.sym,
                args=self.args,
                kwargs=self.kwargs,
            )

        else:
            from scipy.sparse import csr_matrix

            size = sum(self.field.fieldsizes)
            self.results.stiffness = csr_matrix((size, size))

        return self.results.stiffness



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import warnings

import numpy as np

from ..math._math import linsteps
from ..math._tensor import det


class PlotMaterial:
    "Plot force-stretch curves of constitutive material formulations."

    def evaluate(self):
        """Evaluate normal force per undeformed area vs. stretch curves for the
        elementary homogeneous incompressible deformations uniaxial tension/compression,
        planar shear and biaxial tension. A load case is not included if its array of
        stretches  (attribute ``ux``, ``ps`` or ``bx``) is None.

        Returns
        -------
        list of 3-tuple
            List with 3-tuple of stretch and force arrays and the label string for each
            load case.
        """

        data = []

        if self.ux is not None:
            data.append(self.uniaxial())

        if self.ps is not None:
            data.append(self.planar())

        if self.bx is not None:
            data.append(self.biaxial())

        return data

    def plot(self, ax=None, show_title=True, show_kwargs=True, **kwargs):
        """Plot normal force per undeformed area vs. stretch curves for the elementary
        homogeneous incompressible deformations uniaxial tension/compression, planar
        shear and biaxial tension."""

        import matplotlib.pyplot as plt

        if ax is None:
            fig, ax = plt.subplots()

        data = self.evaluate()
        for stretch, force, label in data:
            ax.plot(stretch, force, label=label, **kwargs)

        ax.set_xlabel(r"Stretch $l/L$ $\rightarrow$")
        ax.set_ylabel("Normal force per undeformed area" + r" $N/A$ $\rightarrow$")
        ax.legend()

        xlim = ax.get_xlim()
        ylim = ax.get_ylim()

        ax.plot(xlim, [0, 0], "black")
        ax.plot([1, 1], ylim, "black")

        ax.set_xlim(xlim)
        ax.set_ylim(ylim)

        if show_title:
            title = self.umat.__class__.__name__
            if hasattr(self.umat, "fun"):
                fun = self.umat.fun
                label = [fun.__class__.__name__]
                if callable(fun):
                    label = [name.title() for name in fun.__name__.split("_")]
                title += " (" + " ".join(label) + ")"
            fig.suptitle(title)

        if show_kwargs:
            if hasattr(self.umat, "kwargs"):
                ax.set_title(
                    ", ".join(
                        [f"{key}={value}" for key, value in self.umat.kwargs.items()]
                    ),
                    fontdict=dict(fontsize="small"),
                )
        return ax


class ViewMaterial(PlotMaterial):
    """Create views on normal force per undeformed area vs. stretch curves for the
    elementary homogeneous deformations uniaxial tension/compression, planar shear and
    biaxial tension of a given isotropic material formulation.

    Parameters
    ----------
    umat : class
        A class with methods for the gradient and hessian of the strain energy density
        function w.r.t. the deformation gradient. See :class:`~felupe.Material` for
        further details.
    ux : ndarray, optional
        Array with stretches for uniaxial tension/compression. Default is
        ``linsteps([0.7, 2.5], num=36)``.
    ps : ndarray, optional
        Array with stretches for planar shear. Default is
        ``linsteps([1.0, 2.5], num=30)``.
    bx : ndarray, optional
        Array with stretches for equi-biaxial tension. Default is
        ``linsteps([1.0, 1.75], num=15)```.
    statevars : ndarray or None, optional
        Array with state variables (default is None). If None, the state variables are
        assumed to be initially zero.

    Examples
    --------
    >>> import felupe as fem
    >>>
    >>> umat = fem.OgdenRoxburgh(fem.NeoHooke(mu=1, bulk=2), r=3, m=1, beta=0)
    >>> view = fem.ViewMaterial(
    >>>     umat,
    >>>     ux=fem.math.linsteps([1, 1.5, 1, 2, 1, 2.5, 1], num=15),
    >>>     ps=None,
    >>>     bx=None,
    >>> )
    >>> ax = view.plot(show_title=True, show_kwargs=True)

    .. image:: images/umat.png
       :width: 400px

    """

    def __init__(
        self,
        umat,
        ux=linsteps([0.7, 2.5], num=36),
        ps=linsteps([1, 2.5], num=30),
        bx=linsteps([1, 1.75], num=15),
        statevars=None,
    ):
        self.umat = umat
        self.ux = ux
        self.ps = ps
        self.bx = bx
        self.statevars_included = np.prod(self.umat.x[-1].shape) > 0
        self.statevars = statevars

    def uniaxial(self, stretches=None):
        """Normal force per undeformed area vs. stretch curve for a uniaxial
        deformation.

        Parameters
        ----------
        stretches : ndarray or None, optional
            Array with stretches at which the forces are evaluated (default is None). If
            None, the stretches from initialization are used.

        Returns
        -------
        tuple of ndarray
            3-tuple with array of stretches and array of forces and the label.
        """

        if stretches is None:
            stretches = self.ux

        λ1 = stretches
        λ2 = λ3 = 1 / np.sqrt(λ1)
        eye = np.eye(3).reshape(3, 3, 1, 1)

        def fun(λ3):
            λ2 = λ3
            F = eye * np.array([λ1, λ2, λ3]).reshape(1, 3, 1, -1)
            if self.statevars_included:
                if self.statevars is None:
                    self.statevars = np.zeros((*self.umat.x[-1].shape, 1, 1))
                P = np.zeros_like(F)
                for increment, defgrad in enumerate(F.T):
                    P[..., [increment]], statevars = self.umat.gradient(
                        [F[..., [increment]], self.statevars]
                    )
            else:
                P, statevars = self.umat.gradient([F, None])
            return P[2, 2].ravel()

        from scipy.optimize import root

        λ2 = λ3 = root(fun, λ3).x
        F = eye * np.array([λ1, λ2, λ3]).reshape(1, 3, 1, -1)

        if self.statevars_included:
            if self.statevars is None:
                self.statevars = np.zeros((*self.umat.x[-1].shape, 1, 1))
            P = np.zeros_like(F)
            for increment, defgrad in enumerate(F.T):
                P[..., [increment]], self.statevars = self.umat.gradient(
                    [F[..., [increment]], self.statevars]
                )
        else:
            P, self.statevars = self.umat.gradient([F, None])

        valid = det(F) > np.sqrt(np.finfo(float).eps)
        if not np.all(valid):
            warnings.warn("Uniaxial data with volume ratio det(F) <= 0 included.")
            P[0, 0][~valid] = np.nan

        return λ1, P[0, 0].ravel(), "Uniaxial"

    def planar(self, stretches=None):
        """Normal force per undeformed area vs. stretch curve for a planar shear
        incompressible deformation.

        Parameters
        ----------
        stretches : ndarray or None, optional
            Array with stretches at which the forces are evaluated (default is None). If
            None, the stretches from initialization are used.

        Returns
        -------
        tuple of ndarray
            3-tuple with array of stretches and array of forces and the label.
        """

        if stretches is None:
            stretches = self.ps

        λ1 = stretches
        λ2 = np.ones_like(λ1)
        λ3 = 1 / λ1
        eye = np.eye(3).reshape(3, 3, 1, 1)

        def fun(λ3):
            F = eye * np.array([λ1, λ2, λ3]).reshape(1, 3, 1, -1)
            if self.statevars_included:
                if self.statevars is None:
                    self.statevars = np.zeros((*self.umat.x[-1].shape, 1, 1))
                P = np.zeros_like(F)
                for increment, defgrad in enumerate(F.T):
                    P[..., [increment]], statevars = self.umat.gradient(
                        [F[..., [increment]], self.statevars]
                    )
            else:
                P, statevars = self.umat.gradient([F, None])
            return P[2, 2].ravel()

        from scipy.optimize import root

        λ3 = root(fun, λ3).x
        F = eye * np.array([λ1, λ2, λ3]).reshape(1, 3, 1, -1)

        if self.statevars_included:
            if self.statevars is None:
                self.statevars = np.zeros((*self.umat.x[-1].shape, 1, 1))
            P = np.zeros_like(F)
            for increment, defgrad in enumerate(F.T):
                P[..., [increment]], self.statevars = self.umat.gradient(
                    [F[..., [increment]], self.statevars]
                )
        else:
            P, self.statevars = self.umat.gradient([F, None])

        valid = det(F) > np.sqrt(np.finfo(float).eps)
        if not np.all(valid):
            warnings.warn("Planar Shear data with volume ratio det(F) <= 0 included.")
            P[0, 0][~valid] = np.nan

        return λ1, P[0, 0].ravel(), "Planar Shear"

    def biaxial(self, stretches=None):
        """Normal force per undeformed area vs. stretch curve for a equi-biaxial
        incompressible deformation.

        Parameters
        ----------
        stretches : ndarray or None, optional
            Array with stretches at which the forces are evaluated (default is None). If
            None, the stretches from initialization are used.

        Returns
        -------
        tuple of ndarray
            3-tuple with array of stretches and array of forces and the label.
        """

        if stretches is None:
            stretches = self.bx

        λ1 = λ2 = stretches
        λ3 = 1 / λ1**2
        eye = np.eye(3).reshape(3, 3, 1, 1)

        def fun(λ3):
            F = eye * np.array([λ1, λ2, λ3]).reshape(1, 3, 1, -1)
            if self.statevars_included:
                if self.statevars is None:
                    self.statevars = np.zeros((*self.umat.x[-1].shape, 1, 1))
                P = np.zeros_like(F)
                for increment, defgrad in enumerate(F.T):
                    P[..., [increment]], statevars = self.umat.gradient(
                        [F[..., [increment]], self.statevars]
                    )
            else:
                P, self.statevars = self.umat.gradient([F, None])
            return P[2, 2].ravel()

        from scipy.optimize import root

        λ3 = root(fun, λ3).x
        F = eye * np.array([λ1, λ2, λ3]).reshape(1, 3, 1, -1)

        if self.statevars_included:
            if self.statevars is None:
                self.statevars = np.zeros((*self.umat.x[-1].shape, 1, 1))
            P = np.zeros_like(F)
            for increment, defgrad in enumerate(F.T):
                P[..., [increment]], self.statevars = self.umat.gradient(
                    [F[..., [increment]], self.statevars]
                )
        else:
            P, self.statevars = self.umat.gradient([F, None])

        valid = det(F) > np.sqrt(np.finfo(float).eps)
        if not np.all(valid):
            warnings.warn("Biaxial data with volume ratio det(F) <= 0 included.")
            P[0, 0][~valid] = np.nan

        return λ1, P[0, 0].ravel(), "Biaxial"


class ViewMaterialIncompressible(PlotMaterial):
    """Create views on normal force per undeformed area vs. stretch curves for the
    elementary homogeneous incompressible deformations uniaxial tension/compression,
    planar shear and biaxial tension of a given isotropic material formulation.

    Parameters
    ----------
    umat : class
        A class with methods for the gradient and hessian of the strain energy density
        function w.r.t. the deformation gradient. See :class:`~felupe.Material` for
        further details.
    ux : ndarray, optional
        Array with stretches for incompressible uniaxial tension/compression. Default is
        ``linsteps([0.7, 2.5], num=36)``.
    ps : ndarray, optional
        Array with stretches for incompressible planar shear. Default is
        ``linsteps([1, 2.5], num=30)``.
    bx : ndarray, optional
        Array with stretches for incompressible equi-biaxial tension. Default is
        ``linsteps([1, 1.75], num=15)``.
    statevars : ndarray or None, optional
        Array with state variables (default is None). If None, the state variables are
        assumed to be initially zero.

    Examples
    --------
    >>> import felupe as fem
    >>>
    >>> umat = fem.Hyperelastic(fem.extended_tube, Gc=0.2, Ge=0.2, beta=0.2, delta=0.1)
    >>> preview = fem.ViewMaterialIncompressible(umat)
    >>> ax = preview.plot(show_title=True, show_kwargs=True)

    .. image:: images/preview_hyperelastic.png
       :width: 400px

    >>> umat = fem.OgdenRoxburgh(fem.NeoHooke(mu=1), r=3, m=1, beta=0)
    >>> view = fem.ViewMaterialIncompressible(
    >>>     umat,
    >>>     ux=fem.math.linsteps([1, 1.5, 1, 2, 1, 2.5, 1], num=15),
    >>>     ps=None,
    >>>     bx=None,
    >>> )
    >>> ax = view.plot(show_title=True, show_kwargs=True)

    .. image:: images/umat_incompressible.png
       :width: 400px

    """

    def __init__(
        self,
        umat,
        ux=linsteps([0.7, 2.5], num=36),
        ps=linsteps([1, 2.5], num=30),
        bx=linsteps([1, 1.75], num=15),
        statevars=None,
    ):
        self.umat = umat
        self.ux = ux
        self.ps = ps
        self.bx = bx
        self.statevars_included = np.prod(self.umat.x[-1].shape) > 0
        self.statevars = statevars

    def uniaxial(self, stretches=None):
        """Normal force per undeformed area vs. stretch curve for a uniaxial
        incompressible deformation.

        Parameters
        ----------
        stretches : ndarray or None, optional
            Array with stretches at which the forces are evaluated (default is None). If
            None, the stretches from initialization are used.

        Returns
        -------
        tuple of ndarray
            3-tuple with array of stretches and array of forces and the label.
        """

        if stretches is None:
            stretches = self.ux

        λ1 = stretches
        λ2 = λ3 = 1 / np.sqrt(λ1)
        eye = np.eye(3).reshape(3, 3, 1, 1)
        F = eye * np.array([λ1, λ2, λ3]).reshape(1, 3, 1, -1)

        if self.statevars_included:
            if self.statevars is None:
                self.statevars = np.zeros((*self.umat.x[-1].shape, 1, 1))
            P = np.zeros_like(F)
            for increment, defgrad in enumerate(F.T):
                P[..., [increment]], self.statevars = self.umat.gradient(
                    [F[..., [increment]], self.statevars]
                )
        else:
            P, self.statevars = self.umat.gradient([F, None])

        return λ1, (P[0, 0] - λ3 / λ1 * P[2, 2]).ravel(), "Uniaxial (Incompressible)"

    def planar(self, stretches=None):
        """Normal force per undeformed area vs. stretch curve for a planar shear
        incompressible deformation.

        Parameters
        ----------
        stretches : ndarray or None, optional
            Array with stretches at which the forces are evaluated (default is None). If
            None, the stretches from initialization are used.

        Returns
        -------
        tuple of ndarray
            3-tuple with array of stretches and array of forces and the label.
        """

        if stretches is None:
            stretches = self.ps

        λ1 = stretches
        λ2 = np.ones_like(λ1)
        λ3 = 1 / λ1
        eye = np.eye(3).reshape(3, 3, 1, 1)
        F = eye * np.array([λ1, λ2, λ3]).reshape(1, 3, 1, -1)

        if self.statevars_included:
            if self.statevars is None:
                self.statevars = np.zeros((*self.umat.x[-1].shape, 1, 1))
            P = np.zeros_like(F)
            for increment, defgrad in enumerate(F.T):
                P[..., [increment]], self.statevars = self.umat.gradient(
                    [F[..., [increment]], self.statevars]
                )
        else:
            P, self.statevars = self.umat.gradient([F, None])

        return (
            λ1,
            (P[0, 0] - λ3 / λ1 * P[2, 2]).ravel(),
            "Planar Shear (Incompressible)",
        )

    def biaxial(self, stretches=None):
        """Normal force per undeformed area vs. stretch curve for a equi-biaxial
        incompressible deformation.

        Parameters
        ----------
        stretches : ndarray or None, optional
            Array with stretches at which the forces are evaluated (default is None). If
            None, the stretches from initialization are used.

        Returns
        -------
        tuple of ndarray
            3-tuple with array of stretches and array of forces and the label.
        """

        if stretches is None:
            stretches = self.bx

        λ1 = λ2 = stretches
        λ3 = 1 / λ1**2
        eye = np.eye(3).reshape(3, 3, 1, 1)
        F = eye * np.array([λ1, λ2, λ3]).reshape(1, 3, 1, -1)

        if self.statevars_included:
            if self.statevars is None:
                self.statevars = np.zeros((*self.umat.x[-1].shape, 1, 1))
            P = np.zeros_like(F)
            for increment, defgrad in enumerate(F.T):
                P[..., [increment]], self.statevars = self.umat.gradient(
                    [F[..., [increment]], self.statevars]
                )
        else:
            P, self.statevars = self.umat.gradient([F, None])

        return λ1, (P[0, 0] - λ3 / λ1 * P[2, 2]).ravel(), "Biaxial (Incompressible)"



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from copy import deepcopy

import numpy as np

from ..math import identity
from ..math import sym as symmetric
from ._container import FieldContainer
from ._indices import Indices


class Field:
    r"""A Field on points of a :class:`~felupe.Region` with dimension ``dim`` and
    initial point ``values``.

    Parameters
    ----------
    region : Region
        The region on which the field will be created.
    dim : int, optional
        The dimension of the field  (default is 1).
    values : float or array
        A single value for all components of the field or an array of shape
        `(region.mesh.npoints, dim)`. Default is 0.0.
    **kwargs : dict, optional
        Extra class attributes for the field.

    Notes
    -----
    A slice of this field directly accesses the field-values as 1d-array. The
    interpolation method returns the field values evaluated at the numeric integration
    points ``q`` for each cell ``c`` in the region (so-called *trailing axes*).

    ..  math::

        u_{i(qc)} = \hat{u}_{ai}\ h_{a(qc)}

    The gradient method returns the gradient of the field values w.r.t. the
    undeformed mesh point coordinates, evaluated at the integration points of
    all cells in the region.

    ..  math::

        \left( \frac{\partial u_i}{\partial X_J} \right)_{(qc)} =
            \hat{u}_{ai} \left( \frac{\partial h_a}{\partial X_J} \right)_{(qc)}

    Examples
    --------
    >>> import felupe as fem

    >>> mesh = fem.Cube(n=6)
    >>> region = fem.RegionHexahedron(mesh)
    >>> displacement = fem.Field(region, dim=3)

    >>> u = displacement.interpolate()
    >>> dudX = displacement.grad()

    To obtain deformation-related quantities like the right Cauchy-Green deformation
    tensor or the principal stretches, use the math-helpers from FElupe. These
    functions operate on arrays with trailing axes.

    ..  math::

        \boldsymbol{C} = \boldsymbol{F}^T \boldsymbol{F}

    >>> from felupe.math import dot, transpose, eigvalsh, sqrt

    >>> F = displacement.extract(grad=True, add_identity=True)
    >>> C = dot(transpose(F), F)
    >>> λ = sqrt(eigvalsh(C))

    """

    def __init__(self, region, dim=1, values=0.0, **kwargs):
        self.region = region
        self.dim = dim
        self.shape = self.region.quadrature.npoints, self.region.mesh.ncells

        # set optional user-defined attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

        # init values
        if isinstance(values, np.ndarray):
            if values.size == dim:
                shape = (region.mesh.npoints, dim)
                self.values = np.broadcast_to(values.reshape(-1, dim), shape)
            elif len(values) == region.mesh.npoints:
                self.values = values
            else:
                raise ValueError("Wrong shape of values.")

        else:  # scalar value
            self.values = np.ones((region.mesh.npoints, dim)) * values

        eai, ai = self._indices_per_cell(self.region.mesh.cells, dim)
        self.indices = Indices(eai, ai, region, dim)

    def _indices_per_cell(self, cells, dim):
        "Calculate pre-defined indices for sparse matrices."

        # index of cell "c", point "a" and component "i"
        cai = (
            dim * np.repeat(cells, dim) + np.tile(np.arange(dim), cells.size)
        ).reshape(*cells.shape, dim)
        # store indices as (rows, cols) (note: sparse-matrices are always 2d)
        ai = (cai.ravel(), np.zeros_like(cai.ravel()))

        return cai, ai

    def grad(self, sym=False, out=None):
        """Gradient as partial derivative of field values w.r.t. undeformed coordinates,
        evaluated at the integration points of all cells in the region. Optionally, the
        symmetric part the gradient is evaluated.

        Arguments
        ---------
        sym : bool, optional
            Calculate the symmetric part of the gradient (default is False).
        out : None or ndarray, optional
            A location into which the result is stored. If provided, it must have a
            shape that the inputs broadcast to. If not provided or None, a freshly-
            allocated array is returned (default is None).

        Returns
        -------
        array
            Gradient as partial derivative of field values at points w.r.t.
            undeformed coordinates, evaluated at the integration points of
            all cells in the region.
        """

        # gradient dudX_IJpe as partial derivative of field values at points "aI"
        # w.r.t. undeformed coordinates "J" evaluated at quadrature point "q"
        # for each cell "c"
        g = np.einsum(
            "ca...,aJqc->...Jqc",
            self.values[self.region.mesh.cells],
            self.region.dhdX,
            out=out,
        )

        if sym:
            return symmetric(g)
        else:
            return g

    def interpolate(self, out=None):
        """Interpolate field values at points and evaluate them at the integration
        points of all cells in the region.

        Arguments
        ---------
        out : None or ndarray, optional
            A location into which the result is stored. If provided, it must have a
            shape that the inputs broadcast to. If not provided or None, a freshly-
            allocated array is returned (default is None).
        """

        # interpolated field values "aI"
        # evaluated at quadrature point "q"
        # for cell "c"
        return np.einsum(
            "ca...,aqc->...qc",
            self.values[self.region.mesh.cells],
            self.region.h,
            out=None,
        )

    def extract(self, grad=True, sym=False, add_identity=True, out=None):
        """Generalized extraction method which evaluates either the gradient or the
        field values at the integration points of all cells in the region. Optionally,
        the symmetric part of the gradient is evaluated and/or the identity matrix is
        added to the gradient.

        Arguments
        ---------
        grad : bool, optional
            Flag for gradient evaluation (default is True).
        sym : bool, optional
            Flag for symmetric part if the gradient is evaluated (default is False).
        add_identity : bool, optional
            Flag for the addition of the identity matrix
            if the gradient is evaluated (default is True).
        out : None or ndarray, optional
            A location into which the result is stored. If provided, it must have a
            shape that the inputs broadcast to. If not provided or None, a freshly-
            allocated array is returned (default is None).

        Returns
        -------
        array
            (Symmetric) gradient or interpolated field values evaluated at
            the integration points of each cell in the region.
        """

        if grad:
            gr = self.grad(out=out)

            if sym:
                gr = symmetric(gr, out=gr)

            if add_identity:
                gr = np.add(gr, identity(gr), out=gr)

            return gr
        else:
            return self.interpolate(out=out)

    def copy(self):
        "Return a copy of the field."
        return deepcopy(self)

    def fill(self, a):
        "Fill all field values with a scalar value."
        self.values.fill(a)

    def as_container(self):
        "Create a :class:`~felupe.FieldContainer` with the field."

        return FieldContainer([self])

    def __add__(self, newvalues):
        if isinstance(newvalues, np.ndarray):
            field = deepcopy(self)
            field.values += newvalues.reshape(-1, field.dim)
            return field

        elif isinstance(newvalues, Field):
            field = deepcopy(self)
            field.values += newvalues.values
            return field

        else:
            raise TypeError("Unknown type.")

    def __sub__(self, newvalues):
        if isinstance(newvalues, np.ndarray):
            field = deepcopy(self)
            field.values -= newvalues.reshape(-1, field.dim)
            return field

        elif isinstance(newvalues, Field):
            field = deepcopy(self)
            field.values -= newvalues.values
            return field

        else:
            raise TypeError("Unknown type.")

    def __mul__(self, newvalues):
        if isinstance(newvalues, np.ndarray):
            field = deepcopy(self)
            field.values *= newvalues.reshape(-1, field.dim)
            return field

        elif isinstance(newvalues, Field):
            field = deepcopy(self)
            field.values *= newvalues.values
            return field

        else:
            raise TypeError("Unknown type.")

    def __truediv__(self, newvalues):
        if isinstance(newvalues, np.ndarray):
            field = deepcopy(self)
            field.values /= newvalues.reshape(-1, field.dim)
            return field

        elif isinstance(newvalues, Field):
            field = deepcopy(self)
            field.values /= newvalues.values
            return field

        else:
            raise TypeError("Unknown type.")

    def __iadd__(self, newvalues):
        if isinstance(newvalues, np.ndarray):
            self.values += newvalues.reshape(-1, self.dim)
            return self

        elif isinstance(newvalues, Field):
            self.values += newvalues.values
            return self

        else:
            raise TypeError("Unknown type.")

    def __isub__(self, newvalues):
        if isinstance(newvalues, np.ndarray):
            self.values -= newvalues.reshape(-1, self.dim)
            return self

        elif isinstance(newvalues, Field):
            self.values -= newvalues.values
            return self

        else:
            raise TypeError("Unknown type.")

    def __imul__(self, newvalues):
        if isinstance(newvalues, np.ndarray):
            self.values *= newvalues.reshape(-1, self.dim)
            return self

        elif isinstance(newvalues, Field):
            self.values *= newvalues.values
            return self

        else:
            raise TypeError("Unknown type.")

    def __itruediv__(self, newvalues):
        if isinstance(newvalues, np.ndarray):
            self.values /= newvalues.reshape(-1, self.dim)
            return self

        elif isinstance(newvalues, Field):
            self.values /= newvalues.values
            return self

        else:
            raise TypeError("Unknown type.")

    def __getitem__(self, dof):
        """Slice-based access to a 1d-representation of field values by
        a list of degrees of freedom `dof`."""

        return self.values.ravel()[dof]

    def __and__(self, field):
        fields = [field]
        if isinstance(field, FieldContainer):
            fields = field.fields
        elif field is None:
            fields = []

        return FieldContainer([self, *fields])



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from copy import deepcopy

import numpy as np

from ..tools._plot import ViewField


class FieldContainer:
    """A container for fields which holds a list or a tuple of :class:`Field`
    instances.

    Parameters
    ----------
    fields : list or tuple of Field, FieldAxisymmetric or FieldPlaneStrain
        List with fields. The region is linked to the first field.

    Examples
    --------
    >>> import felupe as fem
    >>>
    >>> mesh = fem.Cube(n=3)
    >>> region = fem.RegionHexahedron(mesh)
    >>> region_dual = fem.RegionConstantHexahedron(mesh.dual(points_per_cell=1))
    >>> displacement = fem.Field(region, dim=3)
    >>> pressure = fem.Field(region_dual)
    >>> field = fem.FieldContainer([displacement, pressure])
    >>> field
    <felupe FieldContainer object>
      Number of fields: 2
      Dimension of fields:
        Field: 3
        Field: 1

    A new :class:`~felupe.FieldContainer` is also created by one of the logical-and
    combinations of a :class:`~felupe.Field`, :class:`~felupe.FieldAxisymmetric`,
    :class:`~felupe.FieldPlaneStrain` or :class:`~felupe.FieldContainer`.

    >>> displacement & pressure
    <felupe FieldContainer object>
      Number of fields: 2
      Dimension of fields:
        Field: 3
        Field: 1

    >>> volume_ratio = fem.Field(region_dual)
    >>> field & volume_ratio  # displacement & pressure & volume_ratio
    <felupe FieldContainer object>
      Number of fields: 3
      Dimension of fields:
        Field: 2
        Field: 1
        Field: 1

    See Also
    --------
    felupe.Field : Field on points of a :class:`~felupe.Region` with dimension ``dim``
        and initial point ``values``.
    felupe.FieldAxisymmetric : An axisymmetric :class:`~felupe.Field` on points of a
        two dimensional :class:`~felupe.Region` with dimension ``dim`` (default is 2)
        and initial point ``values`` (default is 0).
    felupe.FieldPlaneStrain : A plane strain :class:`~felupe.Field` on points of a
        two dimensional :class:`~felupe.Region` with dimension ``dim`` (default is 2)
        and initial point ``values`` (default is 0).

    """

    def __init__(self, fields):
        self.fields = fields
        self.region = fields[0].region

        # get sizes of fields and calculate offsets
        self.fieldsizes = [f.indices.dof.size for f in self.fields]
        self.offsets = np.cumsum(self.fieldsizes)[:-1]

    def __repr__(self):
        header = "<felupe FieldContainer object>"
        size = f"  Number of fields: {len(self.fields)}"
        fields_header = "  Dimension of fields:"
        fields = [
            f"    {type(field).__name__}: {field.dim}"
            for a, field in enumerate(self.fields)
        ]

        return "\n".join([header, size, fields_header, *fields])

    def extract(self, grad=True, sym=False, add_identity=True, out=None):
        """Generalized extraction method which evaluates either the gradient or the
        field values at the integration points of all cells in the region. Optionally,
        the symmetric part of the gradient is evaluated and/or the identity matrix is
        added to the gradient.

        Arguments
        ---------
        grad : bool or list of bool, optional
            Flag(s) for gradient evaluation(s). A boolean value is appplied on the first
            field only and all other fields are extracted with ``grad=False``. To
            enable or disable gradients per-field, use a list of boolean values instead
            (default is True).
        sym : bool, optional
            Flag for symmetric part if the gradient is evaluated (default is False).
        add_identity : bool, optional
            Flag for the addition of the identity matrix if the gradient is evaluated
            (default is True).
        out : None or ndarray, optional
            A location into which the result is stored. If provided, it must have a
            shape that the inputs broadcast to. If not provided or None, a freshly-
            allocated array is returned (default is None).

        Returns
        -------
        tuple of ndarray
            (Symmetric) gradient or interpolated field values evaluated at
            the integration points of each cell in the region.
        """

        if isinstance(grad, bool):
            grad = (grad,)

        if out is None:
            out = [None] * len(self.fields)

        grads = np.pad(grad, (0, len(self.fields) - 1))
        return tuple(
            f.extract(g, sym, add_identity=add_identity, out=res)
            for g, f, res in zip(grads, self.fields, out)
        )

    def values(self):
        "Return the field values."
        return tuple(f.values for f in self.fields)

    def copy(self):
        "Return a copy of the field."
        return deepcopy(self)

    def link(self, other_field):
        "Link value array of other field."
        for field, newfield in zip(self.fields, other_field.fields):
            field.values = newfield.values

    def view(self, point_data=None, cell_data=None, cell_type=None):
        """View the field with optional given dicts of point- and cell-data items.

        Parameters
        ----------
        point_data : dict or None, optional
            Additional point-data dict (default is None).
        cell_data : dict or None, optional
            Additional cell-data dict (default is None).
        cell_type : pyvista.CellType or None, optional
            Cell-type of PyVista (default is None).

        Returns
        -------
        felupe.ViewField
            A object which provides visualization methods for
            :class:`felupe.FieldContainer`.

        See Also
        --------
        felupe.ViewField : Visualization methods for :class:`felupe.FieldContainer`.
        """

        return ViewField(
            self, point_data=point_data, cell_data=cell_data, cell_type=cell_type
        )

    def plot(self, *args, **kwargs):
        """Plot the first field of the container.

        See Also
        --------
        felupe.Scene.plot: Plot method of a scene.
        """
        return self.view().plot(*args, **kwargs)

    def screenshot(
        self,
        *args,
        filename="field.png",
        transparent_background=None,
        scale=None,
        **kwargs,
    ):
        """Take a screenshot of the first field of the container.

        See Also
        --------
        pyvista.Plotter.screenshot: Take a screenshot of a PyVista plotter.
        """

        return self.plot(*args, off_screen=True, **kwargs).screenshot(
            filename=filename,
            transparent_background=transparent_background,
            scale=scale,
        )

    def imshow(self, *args, ax=None, **kwargs):
        """Take a screenshot of the first field of the container, show the image data in
        a figure and return the ax.
        """

        if ax is None:
            import matplotlib.pyplot as plt

            fig, ax = plt.subplots()

        ax.imshow(self.screenshot(*args, filename=None, **kwargs))
        ax.set_axis_off()

        return ax

    def __add__(self, newvalues):
        fields = deepcopy(self)
        if len(newvalues) != len(self.fields):
            newvalues = np.split(newvalues, self.offsets)

        for field, dfield in zip(fields, newvalues):
            field += dfield

        return fields

    def __sub__(self, newvalues):
        fields = deepcopy(self)
        if len(newvalues) != len(self.fields):
            newvalues = np.split(newvalues, self.offsets)

        for field, dfield in zip(fields, newvalues):
            field -= dfield

        return fields

    def __mul__(self, newvalues):
        fields = deepcopy(self)
        if len(newvalues) != len(self.fields):
            newvalues = np.split(newvalues, self.offsets)

        for field, dfield in zip(fields, newvalues):
            field *= dfield

        return fields

    def __truediv__(self, newvalues):
        fields = deepcopy(self)
        if len(newvalues) != len(self.fields):
            newvalues = np.split(newvalues, self.offsets)

        for field, dfield in zip(fields, newvalues):
            field /= dfield

        return fields

    def __iadd__(self, newvalues):
        if len(newvalues) != len(self.fields):
            newvalues = np.split(newvalues, self.offsets)
        for field, dfield in zip(self.fields, newvalues):
            field += dfield
        return self

    def __isub__(self, newvalues):
        if len(newvalues) != len(self.fields):
            newvalues = np.split(newvalues, self.offsets)
        for field, dfield in zip(self.fields, newvalues):
            field -= dfield
        return self

    def __imul__(self, newvalues):
        if len(newvalues) != len(self.fields):
            newvalues = np.split(newvalues, self.offsets)
        for field, dfield in zip(self.fields, newvalues):
            field *= dfield
        return self

    def __itruediv__(self, newvalues):
        if len(newvalues) != len(self.fields):
            newvalues = np.split(newvalues, self.offsets)
        for field, dfield in zip(self.fields, newvalues):
            field /= dfield
        return self

    def __getitem__(self, idx):
        "Slice-based access to underlying fields."

        return self.fields[idx]

    def __len__(self):
        "Number of fields inside the container."

        return len(self.fields)

    def __and__(self, field):
        fields = [field]
        if isinstance(field, FieldContainer):
            fields = field.fields
        elif field is None:
            fields = []

        return FieldContainer([*self.fields, *fields])



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from types import SimpleNamespace

import numpy as np

from ._scheme import Scheme


class Triangle(Scheme):
    r"""A quadrature scheme suitable for Triangles of order 1, 2 or 3 on the interval
    :math:`[0, 1]`.

    Parameters
    ----------
    order : int
        The order of the quadrature scheme.

    Notes
    -----

    The approximation is given by

    ..  math::

        \int_A f(x) dA \approx \sum f(x_q) w_q

    with quadrature points :math:`x_q` and corresponding weights :math:`w_q` [1]_ [2]_.

    References
    ----------
    .. [1] K. J. Bathe, Finite element procedures, 2nd ed. K. J. Bathe, Watertown, MA,
       2014.
    .. [2] O. C. Zienkiewicz, R. L. Taylor and J. Z. Zhu, The Finite Element Method: Its
       Basis and Fundamentals, 7th ed., Elsevier, 2013.
    """

    def __init__(self, order: int):
        scheme = SimpleNamespace()

        if order == 1:
            scheme.points = np.ones((1, 2)) / 3
            scheme.weights = np.ones(1) / 2

        elif order == 2:
            a = 1 / 6
            b = 2 / 3
            scheme.points = np.array([[a, a], [b, a], [a, b]])
            scheme.weights = np.ones(3) / 6

        elif order == 3:
            a = 3 / 5
            b = 1 / 5

            c = 1 / 3
            d = 25 / 96
            scheme.points = np.array([[c, c], [a, b], [b, a], [b, b]])
            scheme.weights = np.array([-9 / 32, d, d, d])

        elif order == 5:
            a = 0.1012865073235
            b = 0.7974269853531
            c = 0.4701420641051
            d = 0.0597158717898
            e = 0.3333333333333
            f = 0.1259391805448
            g = 0.1323941527885
            h = 0.225
            scheme.points = np.array(
                [[a, a], [b, a], [a, b], [c, d], [c, c], [d, c], [e, e]]
            )
            scheme.weights = np.array([f, f, f, g, g, g, h]) / 2

        else:
            raise NotImplementedError("order must be 1, 2, 3 or 5.")

        super().__init__(scheme.points, scheme.weights)
